"""Phabricator module tests."""
from unittest import mock

import pytest

from spicerack import phabricator
from spicerack.tests import get_fixture_path


def test_create_phabricator_ok():
    """It should initialize the instance."""
    phab = phabricator.create_phabricator(get_fixture_path('phabricator', 'valid.conf'))
    assert isinstance(phab, phabricator.Phabricator)


def test_create_phabricator_missing_section():
    """It should raise PhabricatorError if the specified section is missing in the bot config file."""
    with pytest.raises(phabricator.PhabricatorError, match='Unable to find section'):
        phabricator.create_phabricator(get_fixture_path('phabricator', 'valid.conf'), section='nonexistent')


def test_create_phabricator_missing_option():
    """It should raise PhabricatorError if any of the mandatory option is missing in the bot config file."""
    with pytest.raises(phabricator.PhabricatorError, match='Unable to find all required options'):
        phabricator.create_phabricator(get_fixture_path('phabricator', 'invalid.conf'))


@mock.patch('spicerack.phabricator.phabricator.Phabricator', side_effect=RuntimeError)
def test_init_client_raise(mocked_phabricator):
    """It should raise PhabricatorError if unable to instantiate the Phabricator client."""
    with pytest.raises(phabricator.PhabricatorError, match='Unable to instantiate Phabricator client'):
        phabricator.create_phabricator(get_fixture_path('phabricator', 'valid.conf'))

    # Values from the phabricator/valid.conf fixture
    mocked_phabricator.assert_called_once_with(  # nosec
        host='https://phabricator.example.com/api/', username='phab-bot', token='api-12345')


class TestPhabricator:
    """Test class for the Phabricator class."""

    @mock.patch('spicerack.phabricator.phabricator.Phabricator')
    def setup_method(self, _, mocked_phabricator):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_phabricator_client = mocked_phabricator()
        self.phab = phabricator.Phabricator(self.mocked_phabricator_client, dry_run=False)
        self.task_comment_transactions = [{'type': 'comment', 'value': 'new message'}]

    def test_task_comment_ok(self):
        """It should update a task on Phabricator."""
        self.phab.task_comment('T12345', 'new message')
        self.mocked_phabricator_client.maniphest.edit.assert_called_once_with(
            objectIdentifier='T12345', transactions=self.task_comment_transactions)

    def test_task_comment_dry_run(self):
        """It should not update a task on Phabricator when in DRY-RUN mode."""
        phab = phabricator.Phabricator(self.mocked_phabricator_client)
        phab.task_comment('T12345', 'new message')
        assert not self.mocked_phabricator_client.maniphest.edit.called

    def test_task_comment_fail(self):
        """It should raise PhabricatorError if the update operation fails."""
        self.mocked_phabricator_client.maniphest.edit.side_effect = RuntimeError
        with pytest.raises(phabricator.PhabricatorError, match='Unable to update Phabricator task T12345'):
            self.phab.task_comment('T12345', 'new message')

        self.mocked_phabricator_client.maniphest.edit.assert_called_once_with(
            objectIdentifier='T12345', transactions=self.task_comment_transactions)



"""Cookbook module tests."""
import os
import shutil

from unittest import mock

import pytest

from spicerack import cookbook, Spicerack
from spicerack.tests import get_fixture_path, require_caplog, SPICERACK_TEST_PARAMS


COOKBOOKS_BASE_PATH = 'spicerack/tests/fixtures/cookbook'
LIST_COOKBOOKS_ALL = """cookbooks
|-- cookbook
|-- group1
|   `-- group1.cookbook1
|-- group2
|   |-- group2.cookbook2
|   |-- group2.subgroup1
|   |   `-- group2.subgroup1.cookbook3
|   `-- group2.zcookbook4
|-- group3
|   |-- group3.argparse
|   |-- group3.argument_parser_raise
|   |-- group3.argument_parser_raise_system_exit_str
|   |-- group3.keyboard_interrupt
|   |-- group3.non_zero_exit
|   |-- group3.raise_exception
|   |-- group3.raise_system_exit_0
|   |-- group3.raise_system_exit_9
|   |-- group3.raise_system_exit_str
|   `-- group3.subgroup3
|       `-- group3.subgroup3.cookbook4
`-- root
"""
LIST_COOKBOOKS_ALL_VERBOSE = """cookbooks
|-- cookbook: Top level cookbook.
|-- group1: Group1 Test Cookbooks.
|   `-- group1.cookbook1: Group1 Cookbook1.
|-- group2: -
|   |-- group2.cookbook2: Group2 Cookbook2.
|   |-- group2.subgroup1: -
|   |   `-- group2.subgroup1.cookbook3: Group2 Subgroup1 Cookbook3.
|   `-- group2.zcookbook4: UNKNOWN (unable to detect title)
|-- group3: -
|   |-- group3.argparse: Group3 argparse.
|   |-- group3.argument_parser_raise: Group3 argument_parser() raise.
|   |-- group3.argument_parser_raise_system_exit_str: Group3 Raise SystemExit('message') in argument_parser().
|   |-- group3.keyboard_interrupt: Group3 Raise KeyboardInterrupt.
|   |-- group3.non_zero_exit: Group3 Non-Zero return code.
|   |-- group3.raise_exception: Group3 Raise Exception.
|   |-- group3.raise_system_exit_0: Group3 Raise SystemExit(0).
|   |-- group3.raise_system_exit_9: Group3 Raise SystemExit(9).
|   |-- group3.raise_system_exit_str: Group3 Raise SystemExit('message').
|   `-- group3.subgroup3: -
|       `-- group3.subgroup3.cookbook4: Group3 Subgroup3 Cookbook4.
`-- root: Top level cookbook.
"""
LIST_COOKBOOKS_GROUP3 = """cookbooks
`-- group3
    |-- group3.argparse
    |-- group3.argument_parser_raise
    |-- group3.argument_parser_raise_system_exit_str
    |-- group3.keyboard_interrupt
    |-- group3.non_zero_exit
    |-- group3.raise_exception
    |-- group3.raise_system_exit_0
    |-- group3.raise_system_exit_9
    |-- group3.raise_system_exit_str
    `-- group3.subgroup3
        `-- group3.subgroup3.cookbook4
"""
LIST_COOKBOOKS_GROUP3_SUBGROUP3 = """cookbooks
`-- group3
    `-- group3.subgroup3
        `-- group3.subgroup3.cookbook4
"""
COOKBOOKS_MENU_TTY = """#--- cookbooks args=[] ---#
[NOTRUN] cookbook: Top level cookbook.
[0/1] group1: Group1 Test Cookbooks.
[0/3] group2: -
[0/10] group3: -
[NOTRUN] root: Top level cookbook.
q - Quit
h - Help
"""
COOKBOOKS_MENU_NOTTY = """#--- cookbooks args=[] ---#
[NOTRUN] cookbook: Top level cookbook.
[0/1] group1: Group1 Test Cookbooks.
[0/3] group2: -
[0/10] group3: -
[NOTRUN] root: Top level cookbook.
q - Quit
h - Help
Not a tty, exiting.
"""
COOKBOOKS_GROUP1_MENU = """#--- Group1 Test Cookbooks. args=[] ---#
[NOTRUN] cookbook1: Group1 Cookbook1.
b - Back to parent menu
h - Help
"""
COOKBOOKS_GROUP2_MENU = """#--- group2 args=[] ---#
[NOTRUN] cookbook2: Group2 Cookbook2.
[0/1] subgroup1: -
[NOTRUN] zcookbook4: UNKNOWN (unable to detect title)
b - Back to parent menu
h - Help
"""
COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN = """[Namespace(argument=None, k=False), False, False]
#--- group2 args=[] ---#
[PASS] cookbook2: Group2 Cookbook2.
[0/1] subgroup1: -
[NOTRUN] zcookbook4: UNKNOWN (unable to detect title)
b - Back to parent menu
h - Help
"""


def test_argument_parser_converts_path():
    """It should convert a path-based cookbook into a module path."""
    argv = ['group1/cookbook1.py']
    args = cookbook.argument_parser().parse_args(argv)
    assert args.cookbook == 'group1.cookbook1'


def test_argument_parser_keeps_module():
    """It should keep a module-based cookbook as it is."""
    argv = ['group1.cookbook1']
    args = cookbook.argument_parser().parse_args(argv)
    assert args.cookbook == argv[0]


def test_argument_parser_accept_empty():
    """With no args it should not fail."""
    args = cookbook.argument_parser().parse_args([])
    assert args.cookbook is None


def test_parse_args_list():
    """Passing -l/--list should keep the cookbook None and set list."""
    argv = ['--list']
    args = cookbook.argument_parser().parse_args(argv)
    assert args.list
    assert args.cookbook is None


def test_main_wrong_instance_config(capsys):
    """If the configuration file has invalid instance_params it should print an error and exit."""
    ret = cookbook.main(['-c', get_fixture_path('config_wrong_overrides.yaml'), 'cookbook'])
    _, err = capsys.readouterr()
    assert ret == 1
    assert 'Unable to instantiate Spicerack, check your configuration' in err


class TestCookbooks:
    """Test class for the Cookbooks class."""

    def setup_method(self):
        """Initialize the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.spicerack = Spicerack(verbose=False, dry_run=False, **SPICERACK_TEST_PARAMS)
        self.spicerack_dry_run = Spicerack(verbose=False, dry_run=True, **SPICERACK_TEST_PARAMS)
        self.spicerack_verbose = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    @pytest.mark.parametrize('path_filter, verbose, expected', (
        (None, False, LIST_COOKBOOKS_ALL),
        (None, True, LIST_COOKBOOKS_ALL_VERBOSE),
        ('group3', False, LIST_COOKBOOKS_GROUP3),
        ('group3.subgroup3', False, LIST_COOKBOOKS_GROUP3_SUBGROUP3),
        ('group3.non_existent', False, ''),
    ))
    def test_cookbooks_str(self, monkeypatch, path_filter, verbose, expected):
        """The string representation of the Cookbooks should print all the cookbooks as a tree, based on the options."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        if path_filter is None and not verbose:  # For the first test ensure there is no __pycache__ directory.
            shutil.rmtree(os.path.join(COOKBOOKS_BASE_PATH, 'cookbooks', '__pycache__'), ignore_errors=True)

        spicerack = self.spicerack
        if verbose:
            spicerack = self.spicerack_verbose
        cookbooks = cookbook.Cookbooks(COOKBOOKS_BASE_PATH, [], spicerack, path_filter=path_filter)
        assert cookbooks.menu.get_tree() == expected

    def test_cookbooks_non_existent(self):
        """The Cookbooks class initialized with an empty path should not collect any cookbook."""
        cookbooks = cookbook.Cookbooks(os.path.join(COOKBOOKS_BASE_PATH, 'non_existent'), [], self.spicerack)
        assert cookbooks.menu.get_tree() == ''

    @require_caplog
    @pytest.mark.parametrize('module, err_messages, absent_err_messages, code, args', (
        ('cookbook', ['START - Cookbook cookbook', 'END (PASS) - Cookbook cookbook (exit_code=0)'], [], 0, []),
        ('cookbook', [], ['START - Cookbook', 'END ('], cookbook.COOKBOOK_NO_PARSER_WITH_ARGS_RETCODE, ['arg1']),
        ('group3.non_zero_exit', ['END (FAIL) - Cookbook group3.non_zero_exit (exit_code=1)'], [], 1, []),
        ('group3.non_zero_exit', ['END (FAIL) - Cookbook group3.non_zero_exit (exit_code=1)'], [], 1, []),
        ('group3.non_existent', ['Unable to find cookbook'], [], cookbook.COOKBOOK_NOT_FOUND_RETCODE, []),
        ('group3.argparse', [], ['START - Cookbook', 'END ('], 0, ['-h']),
        ('group3.invalid_syntax', ['invalid syntax (invalid_syntax.py, line 7)'], [],
         cookbook.COOKBOOK_NOT_FOUND_RETCODE, []),
        ('group3.keyboard_interrupt', ['Ctrl+c pressed'], [], cookbook.COOKBOOK_INTERRUPTED_RETCODE, []),
        ('group3.argument_parser_raise', ['raised while parsing arguments for cookbook'], ['START - Cookbook', 'END ('],
         cookbook.COOKBOOK_PARSE_ARGS_FAIL_RETCODE, []),
        ('group3.argument_parser_raise_system_exit_str', ["SystemExit('argument_parser')"],
         ['START - Cookbook', 'END ('], cookbook.COOKBOOK_PARSE_ARGS_FAIL_RETCODE, []),
        ('group3.raise_exception', ['Exception: Something went wrong'], [], cookbook.COOKBOOK_EXCEPTION_RETCODE, []),
        ('group3.raise_system_exit_0', ['SystemExit(0) raised'], [], 0, []),
        ('group3.raise_system_exit_9', ['SystemExit(9) raised'], [], 9, []),
        ('group3.raise_system_exit_str', ["SystemExit('message') raised"], [], cookbook.COOKBOOK_EXCEPTION_RETCODE, []),
    ))  # pylint: disable=too-many-arguments
    def test_main_execute_cookbook(self, tmpdir, caplog, module, err_messages, absent_err_messages, code, args):
        """Calling execute_cookbook() should intercept any exception raised."""
        config = {'cookbooks_base_dir': COOKBOOKS_BASE_PATH, 'logs_base_dir': tmpdir.strpath}
        with mock.patch('spicerack.cookbook.load_yaml_config', lambda config_dir: config):
            with mock.patch('spicerack.cookbook.Spicerack', return_value=self.spicerack):
                ret = cookbook.main([module] + args)

        assert ret == code
        for message in err_messages:
            assert message in caplog.text
        for message in absent_err_messages:
            assert message not in caplog.text

    @require_caplog
    def test_main_execute_cookbook_invalid_args(self, tmpdir, capsys, caplog):
        """Calling a cookbook with the wrong args should let argparse print its message."""
        config = {'cookbooks_base_dir': COOKBOOKS_BASE_PATH, 'logs_base_dir': tmpdir.strpath}
        with mock.patch('spicerack.cookbook.load_yaml_config', lambda config_dir: config):
            with mock.patch('spicerack.cookbook.Spicerack', return_value=self.spicerack):
                ret = cookbook.main(['group3.argparse', '--invalid'])

        assert ret == 2
        _, err = capsys.readouterr()
        assert 'Argparse: error: unrecognized arguments' in err
        for message in ('START - Cookbook', 'END ('):
            assert message not in caplog.text

    def test_main_execute_dry_run(self, capsys, tmpdir):
        """Calling main() with a cookbook and dry_run mode should execute it and set the dry run mode."""
        config = {'cookbooks_base_dir': COOKBOOKS_BASE_PATH, 'logs_base_dir': tmpdir.strpath}
        with mock.patch('spicerack.cookbook.load_yaml_config', lambda config_dir: config):
            with mock.patch('spicerack.cookbook.Spicerack', return_value=self.spicerack_dry_run):
                ret = cookbook.main(['-d', 'root'])

        assert ret == 0
        _, err = capsys.readouterr()
        assert 'DRY-RUN' in err

    @require_caplog
    def test_main_list(self, tmpdir, capsys, caplog):
        """Calling main() with the -l/--list option should print the available cookbooks."""
        config = {'cookbooks_base_dir': COOKBOOKS_BASE_PATH, 'logs_base_dir': tmpdir.strpath}
        with mock.patch('spicerack.cookbook.load_yaml_config', lambda config_dir: config):
            with mock.patch('spicerack.cookbook.Spicerack', return_value=self.spicerack):
                ret = cookbook.main(['-l'])

        out, _ = capsys.readouterr()

        assert ret == 0
        assert out == LIST_COOKBOOKS_ALL
        lines = ['Failed to import module cookbooks.group3.invalid_syntax: invalid syntax (invalid_syntax.py, line 7)',
                 'Failed to import module cookbooks.group3.invalid_subgroup: invalid syntax (__init__.py, line 2)']
        for line in lines:
            assert line in caplog.text

    def test_cookbooks_menu_status(self, monkeypatch):
        """Calling status on a CookbooksMenu should show the completed and total tasks."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        cookbooks = cookbook.Cookbooks(COOKBOOKS_BASE_PATH, [], self.spicerack)
        menu = cookbooks.get_item(cookbooks.cookbooks_module_prefix)
        assert menu.status == '0/16'

    def test_cookbooks_menu_status_done(self, monkeypatch):
        """Calling status on a CookbooksMenu with all tasks completed should return DONE."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        cookbooks = cookbook.Cookbooks(COOKBOOKS_BASE_PATH, [], self.spicerack, path_filter='group1')
        menu = cookbooks.get_item('.'.join((cookbooks.cookbooks_module_prefix, 'group1')))
        assert menu.status == '0/1'
        item = cookbooks.get_item('.'.join((cookbooks.cookbooks_module_prefix, 'group1', 'cookbook1')))
        item.status = item.success
        assert menu.status == 'DONE'

    @pytest.mark.parametrize('tty, answer, output', (
        (False, 'q', COOKBOOKS_MENU_NOTTY),
        (True, 'q', COOKBOOKS_MENU_TTY),
        (True, KeyboardInterrupt, COOKBOOKS_MENU_TTY + 'QUIT\n'),
        (True, ['', 'q'], COOKBOOKS_MENU_TTY + COOKBOOKS_MENU_TTY),
        (True, ['invalid', 'q'], COOKBOOKS_MENU_TTY + '==> Invalid input <==\n' + COOKBOOKS_MENU_TTY),
        (True, ['group1', 'b', 'q'], COOKBOOKS_MENU_TTY + COOKBOOKS_GROUP1_MENU + COOKBOOKS_MENU_TTY),
        (True, ['group2', 'cookbook2', 'b', 'q'],
         COOKBOOKS_MENU_TTY + COOKBOOKS_GROUP2_MENU + COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN + COOKBOOKS_MENU_TTY.replace(
            '[0/3] group2', '[1/3] group2')),
        (True, ['group2', 'cookbook2 --argument value', 'b', 'q'],
         COOKBOOKS_MENU_TTY + COOKBOOKS_GROUP2_MENU
         + COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN.replace('argument=None', "argument='value'")
         + COOKBOOKS_MENU_TTY.replace('[0/3] group2', '[1/3] group2')),
        (True, ['group2 -k', 'cookbook2', 'b', 'q'],
         COOKBOOKS_MENU_TTY + COOKBOOKS_GROUP2_MENU.replace('[]', "['-k']")
         + COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN.replace('[]', "['-k']").replace('k=False', 'k=True')
         + COOKBOOKS_MENU_TTY.replace('[0/3] group2', '[1/3] group2')),
        (True, 'h', COOKBOOKS_MENU_TTY
         + cookbook.COOKBOOKS_MENU_HELP_MESSAGE.format(statuses=cookbook.Cookbook.statuses) + '\n'
         + COOKBOOKS_MENU_TTY),
    ))
    @mock.patch('spicerack.cookbook.sys.stdout.isatty')
    @mock.patch('builtins.input')    # pylint: disable=too-many-arguments
    def test_cookbooks_main_menu(self, mocked_input, mocked_tty, tty, answer, output, capsys, tmpdir):
        """Calling main() with a menu should show the menu and allow to interact with it."""
        mocked_tty.return_value = tty
        mocked_input.side_effect = answer
        config = {'cookbooks_base_dir': COOKBOOKS_BASE_PATH, 'logs_base_dir': tmpdir.strpath}
        with mock.patch('spicerack.cookbook.load_yaml_config', lambda config_dir: config):
            with mock.patch('spicerack.cookbook.Spicerack', return_value=self.spicerack):
                ret = cookbook.main([])

        out, _ = capsys.readouterr()
        assert ret == 0
        assert out == output



"""Group3 argument_parser() raise."""
__title__ = __doc__


def argument_parser():
    """As required by spicerack.cookbook."""
    raise RuntimeError('argument_parser() raise')


def run(_args, _spicerack):
    """As required by spicerack.cookbook."""



"""Icinga module."""
import json
import logging
import time

from contextlib import contextmanager
from datetime import timedelta
from typing import Dict, Iterator, List, Mapping, Optional, Sequence

from cumin.transports import Command

from spicerack.administrative import Reason
from spicerack.exceptions import SpicerackError
from spicerack.remote import RemoteHosts
from spicerack.typing import TypeHosts


DOWNTIME_COMMAND: str = 'icinga-downtime -h "{hostname}" -d {duration} -r {reason}'
ICINGA_DOMAIN: str = 'icinga.wikimedia.org'
MIN_DOWNTIME_SECONDS: int = 60  # Minimum time in seconds the downtime can be set
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class IcingaError(SpicerackError):
    """Custom exception class for errors of this module."""


class IcingaStatusParseError(IcingaError):
    """Custom exception class for errors while parsing the Icinga status."""


class IcingaStatusNotFoundError(IcingaError):
    """Custom exception class for errors while parsing the Icinga status."""


class HostsStatus(dict):
    """Represent the status of all Icinga checks for a set of hosts."""

    @property
    def optimal(self) -> bool:
        """Returns :py:data:`True` if all the hosts are optimal, :py:data:`False` otherwise.

        Returns:
            bool: whether all hosts are optimal.

        """
        return all(status.optimal for status in self.values())

    @property
    def non_optimal_hosts(self) -> List[str]:
        """Return the list of hostnames that are not in an optimal state.

        They can either not being up and running or have at least one failed service.

        Returns:
            list: a list of strings with the hostnames.

        """
        return [hostname for hostname, status in self.items() if not status.optimal]

    @property
    def failed_services(self) -> Dict[str, List[str]]:
        """Return the list of service names that are failing for each host that has at least one.

        Returns:
            dict: a dict with hostnames as keys and list of failing service name strings as values.

        """
        return {status.name: status.failed_services for status in self.values() if not status.optimal}

    @property
    def failed_hosts(self) -> List[str]:
        """Return the list of hostnames that are not up and running. They can either be down or unreachable.

        Returns:
            list: the list of strings with the hostnames.

        """
        return [status.name for status in self.values() if status.state != HostStatus.STATE_UP]


class HostStatus:
    """Represent the status of all Icinga checks for a single host."""

    STATE_UP = 'UP'
    """:py:class:`str`: the Icinga value for a host that is up and running. The other values for the Icinga host state
    are ``DOWN`` and ``UNREACHABLE``."""

    def __init__(self, *, name: str, state: str, optimal: bool, failed_services: Sequence[Mapping]):
        """Initialize the instance.

        Arguments:
            name (str): the hostname.
            state (str): the Icinga state for the host, one of ``UP``, ``DOWN``, UNREACHABLE``.
            optimal (bool): whether the host is in optimal state (all green).
            failed_services (list): a list of dictionaries representing the failed services.

        """
        self.name = name
        self.state = state
        self.optimal = optimal
        self.failed_services_raw = failed_services  # TODO: could be improved creating a ServiceStatus class

    @property
    def failed_services(self) -> List[str]:
        """Return the list of service names that are failing.

        Returns:
            list: a list of strings with the check names.

        """
        return [service['name'] for service in self.failed_services_raw]


class Icinga:
    """Class to interact with the Icinga server."""

    def __init__(self, icinga_host: RemoteHosts, *, config_file: str = '/etc/icinga/icinga.cfg') -> None:
        """Initialize the instance.

        Todo:
            Add an IcingaHosts class that performs actions for a given set of hosts instead of passing them to
            all methods.

        Arguments:
            icinga_host (spicerack.remote.RemoteHosts): the RemoteHosts instance for the Icinga server.
            config_file (str, optional): the path to the Icinga main configuration file.

        """
        self._icinga_host = icinga_host
        self._config_file = config_file
        self._command_file: Optional[str] = None

    @property
    def command_file(self) -> Optional[str]:
        """Getter for the command_file property.

        Returns:
            str: the path of the Icinga command file.

        Raises:
            spicerack.icinga.IcingaError: if unable to get the command file path.

        """
        if self._command_file:
            return self._command_file

        try:
            # Get the command_file value in the Icinga configuration.
            command = r"""awk '/^\s*command_file=/{split($0, a, "="); print a[2] }' """ + self._config_file
            for _, output in self._icinga_host.run_sync(command, is_safe=True):
                command_file = output.message().decode().strip()

            if not command_file:
                raise ValueError('Empty or no value found for command_file configuration')

        except (SpicerackError, ValueError) as e:
            raise IcingaError('Unable to read command_file configuration') from e

        self._command_file = command_file
        return self._command_file

    @contextmanager
    def hosts_downtimed(
        self,
        hosts: TypeHosts,
        reason: Reason,
        *,
        duration: timedelta = timedelta(hours=4),
        remove_on_error: bool = False
    ) -> Iterator[None]:
        """Context manager to perform actions while the hosts are downtimed on Icinga.

        Arguments:
            hosts (spicerack.typing.TypeHosts): the set of hostnames or FQDNs to downtime.
            reason (spicerack.administrative.Reason): the reason to set for the downtime on the Icinga server.
            duration (datetime.timedelta, optional): the length of the downtime period.
            remove_on_error: should the downtime be removed even if an exception was raised.

        Yields:
            None: it just yields control to the caller once Icinga has been downtimed and deletes the downtime once
            getting back the control.

        """
        self.downtime_hosts(hosts, reason, duration=duration)
        try:
            yield
        except BaseException:
            if remove_on_error:
                self.remove_downtime(hosts)
            raise
        else:
            self.remove_downtime(hosts)

    def downtime_hosts(
        self,
        hosts: TypeHosts,
        reason: Reason,
        *,
        duration: timedelta = timedelta(hours=4)
    ) -> None:
        """Downtime hosts on the Icinga server for the given time with a message.

        Arguments:
            hosts (spicerack.typing.TypeHosts): the set of hostnames or FQDNs to downtime.
            reason (spicerack.administrative.Reason): the reason to set for the downtime on the Icinga server.
            duration (datetime.timedelta, optional): the length of the downtime period.

        """
        duration_seconds = int(duration.total_seconds())
        if duration_seconds < MIN_DOWNTIME_SECONDS:
            raise IcingaError('Downtime duration must be at least 1 minute, got: {duration}'.format(duration=duration))

        if not hosts:
            raise IcingaError('Got empty hosts list to downtime')

        hostnames = Icinga._get_hostnames(hosts)
        commands = [DOWNTIME_COMMAND.format(hostname=name, duration=duration_seconds, reason=reason.quoted())
                    for name in hostnames]

        logger.info('Scheduling downtime on Icinga server %s for hosts: %s', self._icinga_host, hosts)
        self._icinga_host.run_sync(*commands)

    def recheck_all_services(self, hosts: TypeHosts) -> None:
        """Force recheck of all services associated with a set of hosts.

        Arguments:
            hosts (spicerack.typing.TypeHosts): the set of hostnames or FQDNs to recheck.

        """
        self.host_command('SCHEDULE_FORCED_HOST_SVC_CHECKS', hosts, str(int(time.time())))

    def remove_downtime(self, hosts: TypeHosts) -> None:
        """Remove a downtime from a set of hosts.

        Arguments:
            hosts (spicerack.typing.TypeHosts): the set of hostnames or FQDNs to remove the downtime from.

        """
        self.host_command('DEL_DOWNTIME_BY_HOST_NAME', hosts)

    def host_command(self, command: str, hosts: TypeHosts, *args: str) -> None:
        """Execute a host-specific Icinga command on the Icinga server for a set of hosts.

        Arguments:
            command (str): the Icinga command to execute.
            hosts (spicerack.typing.TypeHosts): the set of hostnames or FQDNs to iterate the command to.
            *args (str): optional positional arguments to pass to the command.

        See Also:
            https://icinga.com/docs/icinga1/latest/en/extcommands2.html

        """
        hostnames = Icinga._get_hostnames(hosts)
        commands = [self._get_command_string(command, hostname, *args) for hostname in hostnames]
        self._icinga_host.run_sync(*commands)

    def get_status(self, hosts: TypeHosts) -> HostsStatus:
        """Get the current status of the given hosts from Icinga.

        Arguments:
            hosts (spicerack.typing.TypeHosts): the set of hostnames or FQDNs to iterate the command to.

        Returns:
            spicerack.icinga.HostsStatus: the instance that represents the status for the given hosts.

        Raises:
            IcingaError: if unable to get the status.
            IcingaStatusParseError: when failing to parse the status.
            IcingaStatusNotFoundError: if a host is not found in the Icinga status.

        """
        # icinga-status exits with non-zero exit code on missing and non-optimal hosts.
        command = Command('/usr/local/bin/icinga-status -j "{hosts}"'.format(hosts=hosts), ok_codes=[])
        for _, output in self._icinga_host.run_sync(command, is_safe=True):  # icinga-status is a read-only script
            json_status = output.message().decode()
            break
        else:
            raise IcingaError('Unable to get the status for the given hosts, no output from icinga-status')

        try:
            status = json.loads(json_status)
        except json.JSONDecodeError as e:
            raise IcingaStatusParseError('Unable to parse Icinga status') from e

        hosts_status = HostsStatus()
        for hostname, host_status in status.items():
            if not host_status:
                raise IcingaStatusNotFoundError('Host {host} was not found in Icinga status'.format(host=hostname))

            hosts_status[hostname] = HostStatus(**host_status)

        return hosts_status

    def _get_command_string(self, *args: str) -> str:
        """Get the Icinga command to execute given the current arguments.

        Arguments:
            *args (str): positional arguments to use to compose the Icinga command string.

        Returns:
            str: the command line to execute on the Icinga host.

        """
        return 'echo -n "[{now}] {args}" > {command_file}'.format(
            now=int(time.time()), args=';'.join(args), command_file=self.command_file)

    @staticmethod
    def _get_hostnames(fqdns: TypeHosts) -> List[str]:
        """Convert FQDNs into hostnames.

        Arguments:
            hosts (spicerack.typing.TypeHosts): an iterable with the list of hostnames to iterate the command for.

        Returns:
            list: the list of hostnames.

        """
        return [fqdn.split('.')[0] for fqdn in fqdns]



"""Interactive module tests."""
import logging
import os
import uuid

from unittest import mock

from spicerack import log
from spicerack.tests import require_caplog


GENERIC_LOG_RECORD = logging.LogRecord('module', logging.DEBUG, '/source/file.py', 1, 'message', [], None)
CUMIN_LOG_RECORD = logging.LogRecord('cumin.module', logging.DEBUG, '/cumin/source/file.py', 1, 'message', [], None)
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def _assert_match_in_tmpdir(match, tmp_dir):
    """Given a match string, assert that it's present in all files in tmp_dir."""
    tmp_dir = str(tmp_dir)  # Newer versions pass a LocalPath, older a string.
    for logfile in os.listdir(tmp_dir):
        with open(os.path.join(tmp_dir, logfile), 'r') as f:
            assert match in f.read()


def test_irc_socket_handler_init():
    """An instance of IRCSocketHandler should set the address for the socket and the user."""
    handler = log.IRCSocketHandler('host', 123, 'user')
    assert handler.addr == ('host', 123)
    assert handler.username == 'user'


@mock.patch('spicerack.log.socket')
def test_irc_socket_handler_emit_ok(mocked_socket):
    """Calling emit() on an IRCSocketHandler instance should send the message to the socket."""
    handler = log.IRCSocketHandler('host', 123, 'user')

    handler.emit(GENERIC_LOG_RECORD)

    mocked_socket.assert_has_calls([
        mock.call.socket().connect(('host', 123)), mock.call.socket().close()], any_order=True)


@mock.patch('spicerack.log.socket')
def test_irc_socket_handler_emit_ko(mocked_socket):
    """If an error occur while calling emit() on an IRCSocketHandler instance, it should call ."""
    handler = log.IRCSocketHandler('host', 123, 'user')
    handler.handleError = mock.MagicMock()
    mocked_socket.socket.side_effect = OSError

    handler.emit(GENERIC_LOG_RECORD)

    assert mock.call.socket().connect(('host', 123)) not in mocked_socket.mock_calls
    assert mock.call.socket().close() not in mocked_socket.mock_calls
    handler.handleError.assert_called_once_with(GENERIC_LOG_RECORD)


def test_cumin_filter_pass():
    """The FilterOutCumin filter() method should let a normal log record pass."""
    log_filter = log.FilterOutCumin()
    ret = log_filter.filter(GENERIC_LOG_RECORD)
    assert ret == 1


def test_cumin_filter_blocks_cumin():
    """The FilterOutCumin filter() method should block a Cumin's log record."""
    log_filter = log.FilterOutCumin()
    ret = log_filter.filter(CUMIN_LOG_RECORD)
    assert ret == 0


@require_caplog
def test_setup_logging_no_irc(tmpdir, caplog):
    """Calling setup_logging() should setup all the handlers of the root logger."""
    log.setup_logging(tmpdir.strpath, 'task', 'user')
    message = str(uuid.uuid4())
    logger.info(message)
    assert message in caplog.text
    _assert_match_in_tmpdir(message, tmpdir.strpath)


@require_caplog
@mock.patch('spicerack.log.socket')
def test_setup_logging_with_irc(mocked_socket, tmpdir, caplog):
    """Calling setup_logging() with host and port should also setup the IRC logger."""
    log.setup_logging(tmpdir.strpath, 'task', 'user', host='host', port=123, dry_run=False)
    message = str(uuid.uuid4())
    log.irc_logger.info(message)

    assert mock.call.socket().connect(('host', 123)) in mocked_socket.mock_calls
    assert message in caplog.text
    _assert_match_in_tmpdir(message, tmpdir.strpath)


@require_caplog
def test_setup_logging_dry_run(capsys, tmpdir, caplog):
    """Calling setup_logging() when in dry run mode should setup all the handlers and the stdout with DEBUG level."""
    log.setup_logging(tmpdir.strpath, 'task', 'user', dry_run=True)
    message = str(uuid.uuid4())
    logger.info(message)

    _, err = capsys.readouterr()
    assert message in err
    assert 'DRY-RUN' in err
    assert message in caplog.text
    _assert_match_in_tmpdir(message, tmpdir.strpath)


@require_caplog
def test_log_task_start(capsys, tmpdir, caplog):
    """Calling log_task_start() should log a START message for the task to both loggers."""
    log.setup_logging(tmpdir.strpath, 'task', 'user')
    message = str(uuid.uuid4())
    log.log_task_start(message)

    logged_message = 'START - ' + message
    _, err = capsys.readouterr()
    assert logged_message in err
    assert logged_message in caplog.text
    _assert_match_in_tmpdir(logged_message, tmpdir.strpath)


@require_caplog
def test_log_task_start_dry_run(capsys, tmpdir, caplog):
    """Calling log_task_start() in dry-run mode should not print a START message for the task to the IRC logger."""
    log.setup_logging(tmpdir.strpath, 'task', 'user', dry_run=True)
    message = str(uuid.uuid4())
    log.log_task_start(message)

    logged_message = 'START - ' + message
    _, err = capsys.readouterr()
    assert logged_message in err
    assert logged_message in caplog.text
    _assert_match_in_tmpdir(logged_message, tmpdir.strpath)


@require_caplog
def test_log_task_end(capsys, tmpdir, caplog):
    """Calling log_task_end() should print an END message for the task."""
    log.setup_logging(tmpdir.strpath, 'task', 'user', dry_run=False)
    message = str(uuid.uuid4())
    log.log_task_end('success', message)

    logged_message = 'END (success) - ' + message
    _, err = capsys.readouterr()
    assert logged_message in err
    assert logged_message in caplog.text
    _assert_match_in_tmpdir(logged_message, tmpdir.strpath)



"""Icinga module tests."""
from datetime import timedelta
from unittest import mock

import pytest

from ClusterShell.MsgTree import MsgTreeElem
from cumin import NodeSet

from spicerack import icinga
from spicerack.administrative import Reason
from spicerack.remote import RemoteHosts
from spicerack.tests import get_fixture_path


def set_mocked_icinga_host_output(mocked_icinga_host, output):
    """Setup the mocked icinga_host return value with the given output."""
    out = MsgTreeElem(output.encode(), parent=MsgTreeElem())
    mocked_icinga_host.run_sync.return_value = iter([(NodeSet('icinga-host'), out)])


class TestIcinga:
    """Test class for the Icinga class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.reason = Reason('Downtime reason', 'user1', 'orchestration-host', task_id='T12345')
        self.mocked_icinga_host = mock.MagicMock(spec_set=RemoteHosts)
        set_mocked_icinga_host_output(self.mocked_icinga_host, '/var/lib/icinga/rw/icinga.cmd')
        self.icinga = icinga.Icinga(self.mocked_icinga_host)

    def test_command_file_ok(self):
        """It should return the command_file setting from the Icinga configuration."""
        assert self.icinga.command_file == '/var/lib/icinga/rw/icinga.cmd'

    @pytest.mark.parametrize('output', (
        '',
        ' ',
    ))
    def test_command_file_raise(self, output):
        """It should raise IcingaError if failing to get the configuration value."""
        set_mocked_icinga_host_output(self.mocked_icinga_host, output)
        with pytest.raises(icinga.IcingaError, match='Unable to read command_file configuration'):
            self.icinga.command_file  # pylint: disable=pointless-statement

    def test_command_file_cached(self):
        """It should return the already cached value of the command_file if accessed again."""
        command_file = self.icinga.command_file
        self.mocked_icinga_host.reset_mock()
        assert self.icinga.command_file == command_file
        assert not self.mocked_icinga_host.called

    @mock.patch('spicerack.icinga.time.time', return_value=1514764800)
    def test_hosts_downtimed(self, mocked_time):
        """It should downtime the hosts on Icinga, yield and delete the downtime once done."""
        hosts = ['host1']
        call = 'icinga-downtime -h "host1" -d 14400 -r {reason}'.format(reason=self.reason.quoted())

        with self.icinga.hosts_downtimed(hosts, self.reason):
            self.mocked_icinga_host.run_sync.assert_called_once_with(call)
            self.mocked_icinga_host.run_sync.reset_mock()

        self.mocked_icinga_host.run_sync.assert_has_calls([
            mock.call('echo -n "[1514764800] DEL_DOWNTIME_BY_HOST_NAME;host1" > /var/lib/icinga/rw/icinga.cmd')])
        assert mocked_time.called

    def test_downtime_is_kept_when_exception_is_raised(self):
        """Downtime should not be removed if an exception is raised."""
        hosts = ['host1']
        call = 'icinga-downtime -h "host1" -d 14400 -r {reason}'.format(reason=self.reason.quoted())

        with pytest.raises(ValueError):
            with self.icinga.hosts_downtimed(hosts, self.reason):
                self.mocked_icinga_host.run_sync.assert_called_once_with(call)
                self.mocked_icinga_host.run_sync.reset_mock()
                raise ValueError()
        assert not self.mocked_icinga_host.run_sync.called

    def test_downtime_is_removed_when_exception_is_raised(self):
        """Downtime should not be removed if an exception is raised."""
        hosts = ['host1']
        call = 'icinga-downtime -h "host1" -d 14400 -r {reason}'.format(reason=self.reason.quoted())

        with pytest.raises(ValueError):
            with self.icinga.hosts_downtimed(hosts, self.reason, remove_on_error=True):
                self.mocked_icinga_host.run_sync.assert_called_once_with(call)
                self.mocked_icinga_host.run_sync.reset_mock()
                raise ValueError()
        assert self.mocked_icinga_host.run_sync.called

    @pytest.mark.parametrize('hosts', (
        ['host1'],
        ['host1', 'host2'],
        NodeSet('host1'),
        NodeSet('host[1-9]'),
    ))
    def test_downtime_hosts_default_params(self, hosts):
        """It should downtime the hosts on the Icinga server with the default params."""
        self.icinga.downtime_hosts(hosts, self.reason)
        calls = [('icinga-downtime -h "{host}" -d 14400 -r {reason}').format(host=host, reason=self.reason.quoted())
                 for host in hosts]
        self.mocked_icinga_host.run_sync.assert_called_once_with(*calls)

    def test_downtime_hosts_custom_duration(self):
        """It should downtime the hosts for the given duration on the Icinga server."""
        self.icinga.downtime_hosts(['host1'], self.reason, duration=timedelta(minutes=30))
        self.mocked_icinga_host.run_sync.assert_called_once_with((
            'icinga-downtime -h "host1" -d 1800 -r {reason}'.format(reason=self.reason.quoted())))

    def test_downtime_hosts_invalid_duration(self):
        """It should raise IcingaError if the duration is too short."""
        with pytest.raises(icinga.IcingaError, match='Downtime duration must be at least 1 minute'):
            self.icinga.downtime_hosts(['host1'], self.reason, duration=timedelta(seconds=59))

    def test_downtime_hosts_no_hosts(self):
        """It should raise IcingaError if there are no hosts to downtime."""
        with pytest.raises(icinga.IcingaError, match='Got empty hosts list to downtime'):
            self.icinga.downtime_hosts([], self.reason)

    @pytest.mark.parametrize('hosts', (
        ['host1'],
        ['host1', 'host2'],
        NodeSet('host1'),
        NodeSet('host[1-9]'),
    ))
    @mock.patch('spicerack.icinga.time.time', return_value=1514764800)
    def test_host_command(self, mocked_time, hosts):
        """It should run the specified command for all the hosts on the Icinga server."""
        self.icinga.host_command('TEST_COMMAND', hosts, 'arg1', 'arg2')
        calls = [
            'echo -n "[1514764800] TEST_COMMAND;{host};arg1;arg2" > /var/lib/icinga/rw/icinga.cmd'.format(host=host)
            for host in hosts]

        self.mocked_icinga_host.run_sync.assert_called_with(*calls)
        assert mocked_time.called

    @mock.patch('spicerack.icinga.time.time', return_value=1514764800)
    def test_recheck_all_services(self, mocked_time):
        """It should force a recheck of all services for the hosts on the Icinga server."""
        self.icinga.recheck_all_services(NodeSet('host1'))
        self.mocked_icinga_host.run_sync.assert_called_with(
            'echo -n "[1514764800] SCHEDULE_FORCED_HOST_SVC_CHECKS;host1;1514764800" > /var/lib/icinga/rw/icinga.cmd')
        assert mocked_time.called

    @mock.patch('spicerack.icinga.time.time', return_value=1514764800)
    def test_remove_downtime(self, mocked_time):
        """It should remove the downtime for the hosts on the Icinga server."""
        self.icinga.remove_downtime(NodeSet('host1'))
        self.mocked_icinga_host.run_sync.assert_called_with(
            'echo -n "[1514764800] DEL_DOWNTIME_BY_HOST_NAME;host1" > /var/lib/icinga/rw/icinga.cmd')
        assert mocked_time.called

    def test_get_status_ok(self):
        """It should parse the JSON payload and return an instance of HostsStatus."""
        with open(get_fixture_path('icinga', 'status_valid.json')) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read())

        status = self.icinga.get_status('host[1-2]')

        assert not status.optimal
        assert status.non_optimal_hosts == ['host2']
        assert status.failed_services == {'host2': ['check_name1', 'check_name2']}
        assert status.failed_hosts == []

    def test_get_status_parse_fail(self):
        """It should raise IcingaStatusParseError if unable to parse the JSON payload."""
        with open(get_fixture_path('icinga', 'status_invalid.json')) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read())

        with pytest.raises(icinga.IcingaStatusParseError, match='Unable to parse Icinga status'):
            self.icinga.get_status('host[1-2]')

    def test_get_status_missing_hosts(self):
        """It should raise IcingaStatusNotFoundError if any host is missing its status."""
        with open(get_fixture_path('icinga', 'status_missing.json')) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read())
        with pytest.raises(icinga.IcingaStatusNotFoundError, match='Host host2 was not found in Icinga status'):
            self.icinga.get_status('host[1-2]')

    def test_get_status_no_output(self):
        """It should raise IcingaError if there is no output from the icinga-status command."""
        self.mocked_icinga_host.run_sync.return_value = iter(())
        with pytest.raises(icinga.IcingaError, match='no output from icinga-status'):
            self.icinga.get_status('host[1-2]')


def _get_hoststatus(hostname, down=False, failed=False):
    """Return an instance of HostStatus suitable for testing based on the parameters.

    Arguments:
        hostname: the name to use for the host.
        down: whether the reported state should be UP or DOWN.
        failed: whether if should have some failed services or not.

    """
    params = {
        'name': hostname,
        'state': 'UP',
        'optimal': True,
        'failed_services': [],
    }

    failed_services = [
        {'host': hostname, 'name': 'check_name1', 'status': {}},
        {'host': hostname, 'name': 'check_name2', 'status': {}},
    ]

    if down:
        params['state'] = 'DOWN'
        params['optimal'] = False

    if failed:
        params['failed_services'] = failed_services
        params['optimal'] = False

    return icinga.HostStatus(**params)


def test_hoststatus_failed_services():
    """It should return the list of check names that are failed."""
    status = _get_hoststatus('hostname', failed=True)
    assert status.failed_services == ['check_name1', 'check_name2']


class TestHostsStatus:
    """Tests for the HostsStatus class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.status = {
            'ok': icinga.HostsStatus(),  # All optimal hosts
            'down': icinga.HostsStatus(),  # Some hosts are down, no hosts have failed services
            'failed': icinga.HostsStatus(),  # All hosts up, but some have failed services
            'down_failed': icinga.HostsStatus(),  # Some hosts are donw, some have failed services
        }

        for i in range(1, 4):
            hostname = 'host{i}'.format(i=i)
            self.status['ok'][hostname] = _get_hoststatus(hostname)
            self.status['down'][hostname] = _get_hoststatus(hostname)
            self.status['failed'][hostname] = _get_hoststatus(hostname)
            self.status['down_failed'][hostname] = _get_hoststatus(hostname)

        for i in range(4, 6):
            hostname = 'host{i}'.format(i=i)
            self.status['down'][hostname] = _get_hoststatus(hostname, down=True)
            self.status['down_failed'][hostname] = _get_hoststatus(hostname, down=True)

        for i in range(6, 8):
            hostname = 'host{i}'.format(i=i)
            self.status['failed'][hostname] = _get_hoststatus(hostname, failed=True)
            self.status['down_failed'][hostname] = _get_hoststatus(hostname, failed=True)

    def test_optimal_ok(self):
        """It should return True if all hosts are optimal."""
        assert self.status['ok'].optimal

    @pytest.mark.parametrize('key', ('down', 'failed', 'down_failed'))
    def test_optimal_ko(self, key):
        """If should return False if any host is not optimal."""
        assert not self.status[key].optimal

    @pytest.mark.parametrize('key, expected', (
        ('ok', []),
        ('down', ['host4', 'host5']),
        ('failed', ['host6', 'host7']),
        ('down_failed', ['host4', 'host5', 'host6', 'host7']),
    ))
    def test_non_optimal_hosts(self, key, expected):
        """It should return the list of hostnames with non optimal status."""
        assert sorted(self.status[key].non_optimal_hosts) == expected

    @pytest.mark.parametrize('key, expected', (
        ('ok', {}),
        ('down', {'host4': [], 'host5': []}),
        ('failed', {'host6': ['check_name1', 'check_name2'], 'host7': ['check_name1', 'check_name2']}),
        ('down_failed', {'host4': [], 'host5': [], 'host6': ['check_name1', 'check_name2'],
                         'host7': ['check_name1', 'check_name2']}),
    ))
    def test_failed_services(self, key, expected):
        """It should return the dictionary with the hostnames as key and the list of failed services as values."""
        assert self.status[key].failed_services == expected

    @pytest.mark.parametrize('key, expected', (
        ('ok', []),
        ('down', ['host4', 'host5']),
        ('failed', []),
        ('down_failed', ['host4', 'host5']),
    ))
    def test_failed_hosts(self, key, expected):
        """It should return the list of hostnames with non optimal status."""
        assert sorted(self.status[key].failed_hosts) == expected



"""Netbox module."""
import logging

from typing import Dict

import pynetbox

from spicerack.exceptions import SpicerackError


NETBOX_DOMAIN: str = 'netbox.wikimedia.org'
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class NetboxError(SpicerackError):
    """General errors raised by this module."""


class NetboxAPIError(NetboxError):
    """Usually a wrapper for pynetbox.RequestError, errors that occur when accessing the API."""


class NetboxHostNotFoundError(NetboxError):
    """Raised when a host is not found for an operation."""


class Netbox:
    """Class which wraps Netbox API operations."""

    def __init__(self, url: str, token: str, *, dry_run: bool = True):
        """Create Netbox instance.

        Arguments:
            url (str): The Netbox top level URL (with scheme and port if necessary)
            token (str): A Netbox API token
            dry_run (bool, optional): set to False to cause writes to Netbox to occur

        """
        self._api = pynetbox.api(url, token=token)
        self._dry_run = dry_run

    @property
    def api(self) -> pynetbox.api:
        """Getter for the Netbox API property.

        Todo:
            When feasible expose instead higher level functionalities.

        Returns:
            pynetbox.api: the Netbox API instance.

        """
        return self._api

    def _fetch_host(self, hostname: str) -> pynetbox.core.response.Record:
        """Fetch a host (dcim.devices) object.

        Arguments:
            hostname (str): the name of the host to fetch

        Returns:
            pynetbox.core.response.Record: the host object from the API

        Raises:
            NetboxAPIError: on API error
            NetboxError: on parameter error
            NetboxHostNotFoundError: if the host is not found

        """
        try:
            host = self._api.dcim.devices.get(name=hostname)
        except pynetbox.RequestError as ex:
            # excepts on other errors
            raise NetboxAPIError('Error retrieving Netbox host') from ex

        if host is None:
            raise NetboxHostNotFoundError
        return host

    def _fetch_virtual_machine(self, hostname: str) -> pynetbox.core.response.Record:
        """Fetch a virrtual machine (virtualization.virtual_machine) object.

        Arguments:
            hostname (str): the name of the host to fetch

        Returns:
            pynetbox.core.response.Record: the host object from the API

        Raises:
            NetboxAPIError: on API error
            NetboxError: on parameter error
            NetboxHostNotFoundError: if the host is not found

        """
        try:
            host = self._api.virtualization.virtual_machines.get(name=hostname)
        except pynetbox.RequestError as ex:
            # excepts on other errors
            raise NetboxAPIError('Error retrieving Netbox VM') from ex

        if host is None:
            raise NetboxHostNotFoundError

        return host

    def put_host_status(self, hostname: str, status: str) -> None:
        """Set the device status.

        Note:
           This method does not operate on virtual machines since they are
           updated automatically from Ganeti into Netbox.

        Arguments:
            hostname (str): the name of the host to operate on
            status (str): A status label or name

        Raises:
            NetboxAPIError: on API error.
            NetboxError: on parameter error.

        """
        status = status.lower()
        host = self._fetch_host(hostname)
        oldstatus = host.status

        if self._dry_run:
            logger.info('Skipping Netbox status update in DRY-RUN mode for host %s %s -> %s',
                        hostname, oldstatus, status)
            return

        host.status = status
        try:
            save_result = host.save()
        except pynetbox.RequestError as ex:
            raise NetboxAPIError(
                'Failed to save Netbox status for host {} {} -> {}'.format(hostname, oldstatus, status)
            ) from ex

        if save_result:
            logger.info('Netbox status updated for host %s %s -> %s', hostname, oldstatus, status)
        else:
            raise NetboxAPIError('Failed to update Netbox status for host {} {} -> {}'.format(
                hostname, oldstatus, status))

    def fetch_host_status(self, hostname: str) -> str:
        """Return the current status of a host as a string.

        Arguments:
            hostname (str): the name of the host status

        Returns:
            str: the normalized status name

        Raises:
            NetboxAPIError: on API error
            NetboxError: on parameter error
            NetboxHostNotFoundError: if the host is not found

        """
        try:
            return self._fetch_host(hostname).status
        except NetboxHostNotFoundError:
            return self._fetch_virtual_machine(hostname).status

    def fetch_host_detail(self, hostname: str) -> Dict:
        """Return a dict containing details about the host.

        Arguments:
            hostname (str): the name of the host to retrieve.

        Returns:
            dict: data about the host

        Raises:
            NetboxAPIError: on API error
            NetboxError: on parameter error
            NetboxHostNotFoundError: if the host is not found

        """
        is_virtual = False
        vm_cluster = 'N/A'
        try:
            host = self._fetch_host(hostname)
        except NetboxHostNotFoundError:
            host = self._fetch_virtual_machine(hostname)
            is_virtual = True
            vm_cluster = host.cluster.name

        ret = host.serialize()
        ret['is_virtual'] = is_virtual
        ret['ganeti_cluster'] = vm_cluster
        return ret



"""Debmonitor module."""
import logging

import requests

from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class DebmonitorError(SpicerackError):
    """Custom exception class for errors of the Debmonitor class."""


class Debmonitor:
    """Class to interact with a Debmonitor website."""

    def __init__(self, host: str, cert: str, key: str, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            host (str): the hostname of the Debmonitor server (without protocol).
            cert (str): the path to the TLS certificate to use to authenticate on Debmonitor.
            key (str): the path to the TLS key to use to authenticate on Debmonitor.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._base_url: str = 'https://{host}'.format(host=host)
        self._cert = cert
        self._key = key
        self._dry_run = dry_run

    def host_delete(self, hostname: str) -> None:
        """Remove a host and all its packages from Debmonitor.

        Arguments:
            host (str): the FQDN of the host to remove from Debmonitor.

        Raises:
            spicerack.debmonitor.DebmonitorError: on failure to delete. It doesn't raise if the host is already absent
                in Debmonitor.

        """
        if self._dry_run:
            logger.debug('Skip removing host %s from Debmonitor in DRY-RUN', hostname)
            return

        url = '{base}/hosts/{host}'.format(base=self._base_url, host=hostname)
        response = requests.delete(url, cert=(self._cert, self._key), timeout=3)

        if response.status_code == requests.codes['no_content']:
            logger.info('Removed host %s from Debmonitor', hostname)
        elif response.status_code == requests.codes['not_found']:
            logger.info('Host %s already missing on Debmonitor', hostname)
        else:
            raise DebmonitorError('Unable to remove host {host} from Debmonitor, got: {code} {msg}'.format(
                host=hostname, code=response.status_code, msg=response.reason))



"""Group3 Raise SystemExit('message') in argument_parser()."""
__title__ = __doc__


def argument_parser():
    """As required by spicerack.cookbook."""
    raise SystemExit('argument_parser')


def run(_args, _spicerack):
    """As required by spicerack.cookbook."""



"""MysqlLegacy module tests."""
from datetime import datetime
from unittest import mock

import pytest

from cumin import Config, NodeSet

from spicerack import mysql_legacy
from spicerack.remote import Remote, RemoteHosts
from spicerack.tests import get_fixture_path, require_caplog
from spicerack.tests.unit.test_remote import mock_cumin


class TestMysqlLegacyRemoteHosts:
    """Test class for the MysqlLegacyRemoteHosts class."""

    @mock.patch('spicerack.remote.transports', autospec=True)
    def setup_method(self, _, mocked_transports):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.config = Config(get_fixture_path('remote', 'config.yaml'))
        self.mocked_transports = mocked_transports
        self.mysql_remote_hosts = mysql_legacy.MysqlLegacyRemoteHosts(RemoteHosts(self.config, NodeSet('host[1-9]'),
                                                                                  dry_run=False))
        self.expected = [(NodeSet('host1'), 'output1')]

    def test_execute(self):
        """Calling run_query() should run the given query in the target hosts."""
        mock_cumin(self.mocked_transports, 0)
        results = self.mysql_remote_hosts.run_query('query1')
        assert [(host, msg.message().decode()) for host, msg in results] == self.expected


class TestMysqlLegacy:
    """MysqlLegacy class tests."""

    @mock.patch('spicerack.remote.transports', autospec=True)
    def setup_method(self, _, mocked_transports):
        """Initialize the test environment for MysqlLegacy."""
        # pylint: disable=attribute-defined-outside-init
        self.config = Config(get_fixture_path('remote', 'config.yaml'))
        self.mocked_transports = mocked_transports
        self.mocked_remote = mock.MagicMock(spec_set=Remote)
        self.mysql = mysql_legacy.MysqlLegacy(self.mocked_remote, dry_run=False)

    def test_get_dbs(self):
        """It should return and instance of MysqlLegacyRemoteHosts for the matching hosts."""
        self.mysql.get_dbs('query')
        self.mocked_remote.query.assert_called_once_with('query')

    @pytest.mark.parametrize('kwargs, query, match', (
        ({}, 'A:db-core', 'db10[01-99],db20[01-99]'),
        ({'datacenter': 'eqiad'}, 'A:db-core and A:eqiad', 'db10[01-99]'),
        ({'section': 's1'}, 'A:db-core and A:db-section-s1',
         'db10[01-10],db20[01-10]'),
        ({'replication_role': 'master'},
         'A:db-core and A:db-role-master',
         'db10[01-11],db20[01-11]'),
        ({'datacenter': 'eqiad', 'section': 's1'},
         'A:db-core and A:eqiad and A:db-section-s1', 'db10[01-10]'),
        ({'datacenter': 'eqiad', 'replication_role': 'master'},
         'A:db-core and A:eqiad and A:db-role-master',
         'db10[01-11]'),
        ({'section': 's1', 'replication_role': 'master'},
         ('A:db-core and A:db-section-s1 and '
          'A:db-role-master'), 'db1001,db2001'),
        ({'datacenter': 'eqiad', 'section': 's1', 'replication_role': 'master'},
         ('A:db-core and A:eqiad and A:db-section-s1 and '
          'A:db-role-master'), 'db1001'),
    ))
    def test_get_core_dbs_ok(self, kwargs, query, match):
        """It should return the right DBs based on the parameters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet(match))
        self.mysql.get_core_dbs(**kwargs)
        self.mocked_remote.query.assert_called_once_with(query)

    @pytest.mark.parametrize('kwargs', (
        {'datacenter': 'invalid'},
        {'section': 'invalid'},
        {'replication_role': 'invalid'},
    ))
    def test_get_core_dbs_fail(self, kwargs):
        """It should raise MysqlLegacyError if called with invalid parameters."""
        message = 'Got invalid {key}'.format(key=list(kwargs.keys())[0])
        with pytest.raises(mysql_legacy.MysqlLegacyError, match=message):
            self.mysql.get_core_dbs(**kwargs)

        assert not self.mocked_remote.query.called

    def test_get_core_dbs_fail_sanity_check(self):
        """It should raise MysqlLegacyError if matching an invalid number of hosts when looking for masters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db1001'))
        with pytest.raises(mysql_legacy.MysqlLegacyError, match='Matched 1 masters, expected 11'):
            self.mysql.get_core_dbs(datacenter='eqiad', replication_role='master')

        assert self.mocked_remote.query.called

    @require_caplog
    @pytest.mark.parametrize('mode, value', (('readonly', b'1'), ('readwrite', b'0')))
    def test_set_core_masters_readonly(self, mode, value, caplog):
        """It should set the masters as read-only/read-write."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db10[01-11]'))
        mock_cumin(self.mocked_transports, 0, retvals=[[('db10[01-11]', value)]])
        getattr(self.mysql, 'set_core_masters_' + mode)('eqiad')
        assert 'SET GLOBAL read_only=' + value.decode() in caplog.text

    @require_caplog
    @pytest.mark.parametrize('readonly, reply', ((True, b'1'), (False, b'0')))
    def test_verify_core_masters_readonly_ok(self, readonly, reply, caplog):
        """Should verify that the masters have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db10[01-11]'))
        mock_cumin(self.mocked_transports, 0, retvals=[[('db10[01-11]', reply)]])
        self.mysql.verify_core_masters_readonly('eqiad', readonly)
        assert 'SELECT @@global.read_only' in caplog.text

    def test_verify_core_masters_readonly_fail(self):
        """Should raise MysqlLegacyError if some masters do not have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db10[01-11]'))
        mock_cumin(self.mocked_transports, 0, retvals=[[('db1001', b'0'), ('db10[02-11]', b'1')]])
        with pytest.raises(mysql_legacy.MysqlLegacyError, match='Verification failed that core DB masters'):
            self.mysql.verify_core_masters_readonly('eqiad', True)

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_core_masters_in_sync_ok(self, mocked_sleep):
        """Should check that all core masters are in sync with the master in the other DC."""
        hosts = NodeSet('db10[01-11]')
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, NodeSet(host)) for host in hosts] * 2
        retvals = [[(host, b'2018-09-06T10:00:00.000000')] for host in hosts]  # first heartbeat
        retvals += [[(host, b'2018-09-06T10:00:01.000000')] for host in hosts]  # second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        self.mysql.check_core_masters_in_sync('eqiad', 'codfw')
        assert not mocked_sleep.called

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_core_masters_in_sync_fail_heartbeat(self, mocked_sleep):
        """Should raise MysqlLegacyError if unable to get the heartbeat from the current master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db1001'))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql_legacy.MysqlLegacyError, match='Unable to get heartbeat from master'):
            self.mysql.check_core_masters_in_sync('eqiad', 'codfw')
        assert not mocked_sleep.called

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_core_masters_in_sync_not_in_sync(self, mocked_sleep):
        """Should raise MysqlLegacyError if a master is not in sync with the one in the other DC."""
        hosts = NodeSet('db10[01-11]')
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, NodeSet(host)) for host in hosts] + [
            RemoteHosts(self.config, NodeSet('db1001'))] * 3
        retvals = [[(host, b'2018-09-06T10:00:00.000000')] for host in hosts]  # first heartbeat
        retvals += [[('db1001', b'2018-09-06T10:00:00.000000')]] * 3  # 3 failed retries of second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        with pytest.raises(mysql_legacy.MysqlLegacyError,
                           match=r'Heartbeat from master db1001 for section .* not yet in sync'):
            self.mysql.check_core_masters_in_sync('eqiad', 'codfw')

        assert mocked_sleep.called

    def test_get_core_masters_heartbeats_wrong_data(self):
        """Should raise MysqlLegacyError if unable to convert the heartbeat into a datetime."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db1001'))
        mock_cumin(self.mocked_transports, 0, retvals=[[('db1001', b'2018-09-06-10:00:00.000000')]])
        with pytest.raises(mysql_legacy.MysqlLegacyError, match='Unable to convert heartbeat'):
            self.mysql.get_core_masters_heartbeats('eqiad', 'codfw')

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_core_masters_heartbeats_fail(self, mocked_sleep):
        """Should raise MysqlLegacyError if unable to get the heartbeat from the master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db1001'))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql_legacy.MysqlLegacyError, match='Unable to get heartbeat from master'):
            self.mysql.check_core_masters_heartbeats('eqiad', 'codfw', {'s1': datetime.utcnow()})

        assert mocked_sleep.called



"""Vulture whitelist to avoid false positives."""


class Whitelist:
    """Helper class that allows mocking Python objects."""

    def __getattr__(self, _):
        """Mocking magic method __getattr__."""
        pass


whitelist_logging = Whitelist()
whitelist_logging.raiseExceptions

# Needed for vulture < 0.27
whitelist_mock = Whitelist()
whitelist_mock.return_value
whitelist_mock.side_effect

whitelist_log = Whitelist()
whitelist_log.IRCSocketHandler.level

whitelist_dnsdisc = Whitelist()
whitelist_dnsdisc.pool
whitelist_dnsdisc.depool

whitelist_mysql = Whitelist()
whitelist_mysql.set_core_masters_readonly
whitelist_mysql.set_core_masters_readwrite

whitelist_remote = Whitelist()
whitelist_remote.execute.worker.commands
whitelist_remote.execute.worker.commands
whitelist_remote.execute.worker.handler
whitelist_remote.execute.worker.success_threshold
whitelist_remote.run_async
whitelist_remote.run_sync

whitelist_tests = Whitelist()
whitelist_tests.unit.test_confctl.TestConfctl.setup_method
whitelist_tests.unit.test_confctl.TestConfctl.setup_method.backend
whitelist_tests.unit.test_confctl.TestConfctl.setup_method.config
whitelist_tests.unit.test_elasticsearch_cluster.pytestmark
whitelist_tests.unit.test_remote.TestRemote.setup_method
whitelist_tests.unit.test_remote.TestRemote.teardown_method



"""Group3 argparse."""
import argparse


__title__ = __doc__


def argument_parser():
    """As required by spicerack.cookbook."""
    return argparse.ArgumentParser('Argparse')


def run(_args, _spicerack):
    """As required by spicerack.cookbook."""



"""Spicerack package."""
import os

from logging import Logger
from socket import gethostname
from typing import Dict, Optional

from pkg_resources import DistributionNotFound, get_distribution

from spicerack import interactive
from spicerack.actions import ActionsDict
from spicerack.administrative import Reason
from spicerack.confctl import Confctl, ConftoolEntity
from spicerack.config import load_ini_config, load_yaml_config
from spicerack.debmonitor import Debmonitor
from spicerack.dns import Dns
from spicerack.dnsdisc import Discovery
from spicerack.elasticsearch_cluster import create_elasticsearch_clusters, ElasticsearchClusters
from spicerack.ganeti import Ganeti
from spicerack.icinga import Icinga, ICINGA_DOMAIN
from spicerack.ipmi import Ipmi
from spicerack.log import irc_logger
from spicerack.management import Management
from spicerack.mediawiki import MediaWiki
from spicerack.mysql import Mysql
from spicerack.mysql_legacy import MysqlLegacy
from spicerack.netbox import Netbox, NETBOX_DOMAIN
from spicerack.phabricator import create_phabricator, Phabricator
from spicerack.prometheus import Prometheus
from spicerack.puppet import get_puppet_ca_hostname, PuppetHosts, PuppetMaster
from spicerack.redis_cluster import RedisCluster
from spicerack.remote import Remote, RemoteHosts


try:
    __version__: str = get_distribution('wikimedia-spicerack').version  # Must be the same used as 'name' in setup.py
    """:py:class:`str`: the version of the current Spicerack module."""
except DistributionNotFound:  # pragma: no cover - this should never happen during tests
    pass  # package is not installed


class Spicerack:
    """Spicerack service locator."""

    def __init__(
        self, *,
        verbose: bool = False,
        dry_run: bool = True,
        cumin_config: str = '/etc/cumin/config.yaml',
        conftool_config: str = '/etc/conftool/config.yaml',
        conftool_schema: str = '/etc/conftool/schema.yaml',
        debmonitor_config: str = '/etc/debmonitor.conf',
        spicerack_config_dir: str = '/etc/spicerack',
        http_proxy: str = ''
    ) -> None:
        """Initialize the service locator for the Spicerack library.

        Arguments:
            verbose (bool, optional): whether to set the verbose mode.
            dry_run (bool, optional): whether this is a DRY-RUN.
            cumin_config (str, optional): the path to Cumin's configuration file.
            conftool_config (str, optional): the path to Conftool's configuration file.
            conftool_schema (str, optional): the path to Conftool's schema file.
            debmonitor_config (str, optional): the path to Debmonitor's INI configuration file. It must have at least
                the following schema::

                    [DEFAULT]
                    server=debmonitor.example.com
                    cert=/etc/debmonitor/ssl/cert.pem
                    key=/etc/debmonitor/ssl/server.key

            spicerack_config_dir (str, optional): the path for the root configuration directory for Spicerack.
                Module-specific configuration will be loaded from `config_dir/module_name/`.
            http_proxy (str, optional): the scheme://url:port of the HTTP proxy to use for external calls.

        """
        # Attributes
        self._verbose = verbose
        self._dry_run = dry_run
        self._http_proxy = http_proxy
        self._cumin_config = cumin_config
        self._conftool_config = conftool_config
        self._conftool_schema = conftool_schema
        self._debmonitor_config = debmonitor_config
        self._spicerack_config_dir = spicerack_config_dir

        self._username = interactive.get_username()
        self._current_hostname = gethostname()
        self._irc_logger = irc_logger
        self._confctl: Optional[Confctl] = None
        self._ipmi: Optional[Ipmi] = None
        self._actions = ActionsDict()

    @property
    def dry_run(self) -> bool:
        """Getter for the ``dry_run`` property.

        Returns:
            bool: True if the DRY-RUN mode is set, False otherwise.

        """
        return self._dry_run

    @property
    def verbose(self) -> bool:
        """Getter for the ``verbose`` property.

        Returns:
            bool: True if the verbose mode is set, False otherwise.

        """
        return self._verbose

    @property
    def username(self) -> str:
        """Getter for the current username.

        Returns:
            str: the name of the effective running user.

        """
        return self._username

    @property
    def config_dir(self) -> str:
        """Getter for Spicerack's configuration file directory.

        Returns:
            str: a filesystem location of configuration files.

        """
        return self._spicerack_config_dir

    @property
    def http_proxy(self) -> str:
        """Getter for the HTTP PROXY to use for external calls.

        Returns:
            str: the scheme://url:port of the proxy.

        """
        return self._http_proxy

    @property
    def requests_proxies(self) -> Optional[Dict[str, str]]:
        """Getter to return the HTTP proxy configuration for the Requests module.

        Returns:
            dict: with the proxies as required by Requests documentation.
            :py:data:`None`: if no HTTP proxy is set.

        See Also:
            https://requests.readthedocs.io/en/master/user/advanced/#proxies

        """
        if not self._http_proxy:
            return None

        return {'http': self._http_proxy, 'https': self._http_proxy}

    @property
    def irc_logger(self) -> Logger:
        """Getter for the ``irc_logger`` property.

        Returns:
            logging.Logger: the logger instance to write to IRC.

        """
        return self._irc_logger

    @property
    def actions(self) -> ActionsDict:
        """Getter for the ``actions`` property.

        Returns:
            spicerack.actions.ActionsDict: a dictionary to log and record cookbook actions.

        """
        return self._actions

    @property
    def icinga_master_host(self) -> RemoteHosts:
        """Getter for the ``icinga_master_host`` property.

        Returns:
            spicerack.remote.RemoteHosts: the instance to execute commands on the Icinga master host.

        """
        return self.remote().query(self.dns().resolve_cname(ICINGA_DOMAIN))

    @property
    def netbox_master_host(self) -> RemoteHosts:
        """Getter for the ``netbox_master_host`` property.

        Returns:
            spicerack.remote.RemoteHosts: the instance to execute commands on the Netbox master host.

        """
        return self.remote().query(self.dns().resolve_cname(NETBOX_DOMAIN))

    def remote(self) -> Remote:
        """Get a Remote instance.

        Returns:
            spicerack.remote.Remote: the Remote instance.

        """
        return Remote(self._cumin_config, dry_run=self._dry_run)

    def confctl(self, entity_name: str) -> ConftoolEntity:
        """Access a Conftool specific entity instance.

        Arguments:
            entity_name (str): the name of a Conftool entity. Possible values: ``node``, ``service``, ``discovery``,
                ``mwconfig``.

        Returns:
            spicerack.confctl.ConftoolEntity: the confctl entity instance.

        """
        if self._confctl is None:
            self._confctl = Confctl(config=self._conftool_config, schema=self._conftool_schema, dry_run=self._dry_run)

        return self._confctl.entity(entity_name)

    def dns(self) -> Dns:  # pylint: disable=no-self-use
        """Get a Dns instance.

        Returns:
            spicerack.dns.Dns: a Dns instance that will use the operating system default namserver(s).

        """
        return Dns()

    def discovery(self, *records: str) -> Discovery:
        """Get a Discovery instance.

        Arguments:
            *records (str): arbitrary positional arguments, each one must be a Discovery DNS record name.

        Returns:
            spicerack.dnsdisc.Discovery: the pre-configured Discovery instance for the given records.

        """
        return Discovery(self.confctl('discovery'), self.remote(), list(records), dry_run=self._dry_run)

    def mediawiki(self) -> MediaWiki:
        """Get a MediaWiki instance.

        Returns:
            spicerack.mediawiki.MediaWiki: the MediaWiki instance.

        """
        return MediaWiki(self.confctl('mwconfig'), self.remote(), self._username, dry_run=self._dry_run)

    def mysql(self) -> Mysql:
        """Get a Mysql instance.

        Returns:
            spicerack.mysql.Mysql: the Mysql instance.

        """
        return Mysql(dry_run=self._dry_run)

    def mysql_legacy(self) -> MysqlLegacy:
        """Get a MysqlLegacy instance.

        Returns:
            spicerack.mysql_legacy.MysqlLegacy: the MysqlLegacy instance.

        """
        return MysqlLegacy(self.remote(), dry_run=self._dry_run)

    def redis_cluster(self, cluster: str) -> RedisCluster:
        """Get a RedisCluster instance.

        Arguments:
            cluster (str): the name of the cluster.

        Returns:
            spicerack.redis_cluster.RedisCluster: the cluster instance.

        """
        return RedisCluster(cluster, os.path.join(self._spicerack_config_dir, 'redis_cluster'), dry_run=self._dry_run)

    def elasticsearch_clusters(self, clustergroup: str) -> ElasticsearchClusters:
        """Get an ElasticsearchClusters instance.

        Arguments:
            clustergroup (str): name of cluster group e.g ``search_eqiad``.

        Returns:
            spicerack.elasticsearch_cluster.ElasticsearchClusters: ElasticsearchClusters instance.

        """
        return create_elasticsearch_clusters(clustergroup, self.remote(), self.prometheus(), dry_run=self._dry_run)

    def admin_reason(self, reason: str, task_id: Optional[str] = None) -> Reason:
        """Get an administrative Reason instance.

        Arguments:
            reason (str): the reason to use to justify an administrative action. See `spicerack.administrative.Reason`
                for all the details.
            task_id (str, optional): the task ID to mention in the reason.

        Returns:
            spicerack.administrative.Reason: the administrative Reason instance.

        """
        return Reason(reason, self._username, self._current_hostname, task_id=task_id)

    def icinga(self) -> Icinga:
        """Get an Icinga instance.

        Returns:
            spicerack.icinga.Icinga: Icinga instance.

        """
        return Icinga(self.icinga_master_host)

    def puppet(self, remote_hosts: RemoteHosts) -> PuppetHosts:  # pylint: disable=no-self-use
        """Get a PuppetHosts instance for the given remote hosts.

        Arguments:
            remote_hosts (spicerack.remote.RemoteHosts): the instance with the target hosts.

        Returns:
            spicerack.puppet.PuppetHosts: the instance to manage Puppet on the target hosts.

        """
        return PuppetHosts(remote_hosts)

    def puppet_master(self) -> PuppetMaster:
        """Get a PuppetMaster instance to manage hosts and certificates from a Puppet master.

        Returns:
            spicerack.puppet.PuppetMaster: the instance to manage Puppet hosts and certificates.

        """
        return PuppetMaster(self.remote().query(get_puppet_ca_hostname()))

    def ipmi(self, *, cached: bool = False) -> Ipmi:
        """Get an Ipmi instance to send remote IPMI commands to management consoles.

        Arguments:
            cached (bool, optional): whether to cache the Ipmi instance and allow to re-use it without re-asking the
                management password. The cached instance will be returned at each future call of this method with
                the cached parameter set to :py:data:`True`.

        Returns:
            spicerack.ipmi.Ipmi: the instance to run ipmitool commands.

        """
        if cached and self._ipmi is not None:
            return self._ipmi

        ipmi = Ipmi(interactive.get_management_password(), dry_run=self._dry_run)
        if cached:
            self._ipmi = ipmi

        return ipmi

    def phabricator(self, bot_config_file: str, section: str = 'phabricator_bot') -> Phabricator:
        """Get a Phabricator instance to interact with a Phabricator website.

        Arguments:
            bot_config_file (str): the path to the configuration file for the Phabricator bot, with the following
                structure::

                    [section_name]
                    host = https://phabricator.example.com/api/
                    username = phab-bot
                    token = api-12345

            section (str, optional): the name of the section of the configuration file where to find the required
                parameters.

        Returns:
            spicerack.phabricator.Phabricator: the instance.

        """
        # Allow to specify the configuration file as opposed to other methods so that different clients can use
        # different Phabricator BOT accounts, potentially with different permissions.
        return create_phabricator(bot_config_file, section=section, dry_run=self._dry_run)

    def prometheus(self) -> Prometheus:  # pylint: disable=no-self-use
        """Get an Prometheus instance.

        Returns:
            spicerack.prometheus.Prometheus: Prometheus instance.

        """
        return Prometheus()

    def debmonitor(self) -> Debmonitor:
        """Get a Debmonitor instance to interact with a Debmonitor website.

        Returns:
            spicerack.debmonitor.Debmonitor: the instance.

        Raises:
            KeyError: if any configuration option is missing.

        """
        options = load_ini_config(self._debmonitor_config).defaults()
        return Debmonitor(options['server'], options['cert'], options['key'], dry_run=self._dry_run)

    def management(self) -> Management:
        """Get a Management instance to interact with the management interfaces.

        Returns:
            spicerack.management.Management: the instance.

        """
        return Management(self.dns())

    def ganeti(self) -> Ganeti:
        """Get an instance to interact with Ganeti.

        Returns:
            spicerack.ganeti.Ganeti: the instance

        Raises:
            KeyError: If the configuration file does not contain the correct keys.

        """
        configuration = load_yaml_config(os.path.join(self._spicerack_config_dir, 'ganeti', 'config.yaml'))

        return Ganeti(configuration['username'], configuration['password'], configuration['timeout'], self.remote())

    def netbox(self, *, read_write: bool = False) -> Netbox:
        """Get a Netbox instance to interact with Netbox's API.

        Arguments:
            read_write (bool, optional): whether to use a read-write token.

        Returns:
            spicerack.netbox.Netbox: the instance

        """
        config = load_yaml_config(os.path.join(self._spicerack_config_dir, 'netbox', 'config.yaml'))
        if read_write and not self._dry_run:
            token = config['api_token_rw']
        else:
            token = config['api_token_ro']

        return Netbox(config['api_url'], token, dry_run=self._dry_run)



"""Log module."""
import logging
import os
import socket

from pathlib import Path
from typing import Optional, Tuple, Union


root_logger = logging.getLogger()  # pylint: disable=invalid-name
irc_logger = logging.getLogger('spicerack_irc_announce')  # pylint: disable=invalid-name


class IRCSocketHandler(logging.Handler):
    """Log handler for logmsgbot on #wikimedia-operation.

    Sends log events to a tcpircbot server for relay to an IRC channel.
    """

    def __init__(self, host: str, port: int, username: str) -> None:
        """Initialize the IRC socket handler.

        Arguments:
            host (str): tcpircbot hostname.
            port (int): tcpircbot listening port.
            username (str): the user to refer in the IRC messages.

        """
        super().__init__()
        self.addr: Tuple[str, int] = (host, port)
        self.username = username
        self.level = logging.INFO

    def emit(self, record: logging.LogRecord) -> None:
        """According to Python logging.Handler interface.

        See https://docs.python.org/3/library/logging.html#handler-objects
        """
        # Stashbot expects !log messages relayed by logmsgbot to have the
        # format: "!log <nick> <msg>". The <nick> is parsed out and used as
        # the label of who actually made the SAL entry.
        message = '!log {user}@{host} {msg}'.format(
            msg=record.getMessage(), user=self.username, host=socket.gethostname())
        sock = None

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(1.0)
            sock.connect(self.addr)
            sock.sendall(message.encode('utf-8'))
        except OSError:
            self.handleError(record)
        finally:
            if sock is not None:
                sock.close()


class FilterOutCumin(logging.Filter):
    """A logging output filter to filter out Cumin's logs."""

    def filter(self, record: logging.LogRecord) -> int:
        """Filter out Cumin's log messages.

        Arguments:
            record (logging.LogRecord):

        Returns:
            int: 0 if the record should be filtered out, non-zero if it should be included. According to Python's
            logging interface, see: https://docs.python.org/3/library/logging.html#filter-objects

        """
        if record.name == 'cumin' or record.name.startswith('cumin.'):
            return 0  # Filter it out

        return 1


def setup_logging(
    base_path: Union[str, Path],
    name: str,
    user: str,
    dry_run: bool = True,
    host: Optional[str] = None,
    port: int = 0
) -> None:
    """Setup the root logger instance.

    Arguments:
        base_path (str, pathlib.Path): the base path where to save the logs.
        name (str): the name of log file to use without extension.
        user (str): the username for the IRC logging.
        dry_run (bool, optional): whether this is a dry-run.
        host (str, optional): the tcpircbot hostname for the IRC logging.
        port (int, optional): the tcpircbot port for the IRC logging.

    """
    logging.raiseExceptions = False
    os.makedirs(base_path, mode=0o755, exist_ok=True)

    # Default INFO logging
    formatter = logging.Formatter(fmt='%(asctime)s {} %(process)d [%(levelname)s] %(message)s'.format(user))
    handler = logging.FileHandler(os.path.join(base_path, '{name}.log'.format(name=name)))
    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)

    # Extended logging for detailed debugging
    formatter_extended = logging.Formatter(
        fmt='%(asctime)s {} %(process)d [%(levelname)s %(filename)s:%(lineno)s in %(funcName)s] %(message)s'
            .format(user))
    handler_extended = logging.FileHandler(os.path.join(base_path, '{name}-extended.log'.format(name=name)))
    handler_extended.setFormatter(formatter_extended)
    handler_extended.setLevel(logging.DEBUG)

    # Stderr logging
    output_handler = logging.StreamHandler()
    if dry_run:
        output_handler.setFormatter(logging.Formatter(fmt='DRY-RUN: %(message)s'))
        output_handler.setLevel(logging.DEBUG)
    else:
        output_handler.setLevel(logging.INFO)
    output_handler.addFilter(FilterOutCumin())

    root_logger.addHandler(handler)
    root_logger.addHandler(handler_extended)
    root_logger.addHandler(output_handler)
    root_logger.setLevel(logging.DEBUG)

    if not dry_run and host is not None and port > 0:
        irc_logger.addHandler(IRCSocketHandler(host, port, user))
        irc_logger.setLevel(logging.INFO)

    # Silence external noisy loggers
    logging.getLogger('urllib3').setLevel(logging.WARNING)
    logging.getLogger('requests').setLevel(logging.WARNING)

    # Elasticsearch lib is very noisy about HTTP level errors
    # ideally, we'd want to keep it at WARNING level for logs
    # sent to file, but ERROR for the console. Since this is
    # non trivial, let's raise level to ERROR for the moment.
    logging.getLogger('elasticsearch').setLevel(logging.ERROR)


def log_task_start(message: str) -> None:
    """Log the start of a task both on the logs and IRC.

    Arguments:
        message (str): the message to be logged.

    """
    irc_logger.info('START - %s', message)


def log_task_end(status: str, message: str) -> None:
    """Log the start of a task both on the logs and IRC.

    Arguments:
        status (str): the final status of the task.
        message (str): the message to be logged.

    """
    irc_logger.info('END (%s) - %s', status, message)



"""Management module tests."""
from unittest import mock

import pytest

from spicerack.constants import ALL_DATACENTERS
from spicerack.dns import Dns, DnsError
from spicerack.management import Management, ManagementError


class TestManagement:
    """Test class for the Managament class."""

    def setup_method(self):
        """Setup the test class."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_dns = mock.Mock(spec_set=Dns)
        self.management = Management(self.mocked_dns)

    def _setup_mocked_dns(self, *args):
        """Set the mocked Dns responses based on the given args."""
        if len(args) == 1:
            self.mocked_dns.resolve_ipv4.side_effect = args[0]
        else:
            self.mocked_dns.resolve_ipv4.side_effect = list(args)

    def test_get_management_internal_ok(self):
        """It should return the management interface for an internal hostname."""
        hostname = 'host1.eqiad.wmnet'
        mgmt = 'host1.mgmt.eqiad.wmnet'
        self._setup_mocked_dns('127.0.0.1')
        assert self.management.get_fqdn(hostname) == mgmt

    def test_get_management_internal_raise(self):
        """It should raise ManagementError for an internal hostname that doesn't resolve."""
        hostname = 'host1.eqiad.wmnet'
        self._setup_mocked_dns(DnsError)
        with pytest.raises(ManagementError, match='Invalid management FQDN'):
            self.management.get_fqdn(hostname)

    def test_get_management_external_known_dc_ok(self):
        """It should return the management interface for an external hostname that matches naming conventions."""
        hostname = 'host2001.example.com'
        mgmt = 'host2001.mgmt.codfw.wmnet'
        self._setup_mocked_dns('127.0.0.1')
        assert self.management.get_fqdn(hostname) == mgmt

    def test_get_management_external_known_dc_raise(self):
        """It should raise ManagementError for an external hostname that matches naming conventions."""
        hostname = 'host2001.example.com'
        self._setup_mocked_dns(DnsError)
        with pytest.raises(ManagementError, match='Unable to find management FQDN for host'):
            self.management.get_fqdn(hostname)

    @pytest.mark.parametrize('dc_index', range(len(ALL_DATACENTERS)))
    def test_get_management_external_guess_ok(self, dc_index):
        """It should find the management interface for an external hostname that doesn't match naming conventions."""
        hostname = 'host1.example.com'
        mgmt = 'host1.mgmt.{dc}.wmnet'.format(dc=ALL_DATACENTERS[dc_index])
        ret_values = [DnsError] * dc_index + ['127.0.0.1']
        self._setup_mocked_dns(*ret_values)
        assert self.management.get_fqdn(hostname) == mgmt

    def test_get_management_external_guess_raise(self):
        """It should raise ManagementError for an external hostname that doesn't match naming conventions."""
        hostname = 'host1.example.com'
        self._setup_mocked_dns(DnsError)
        with pytest.raises(ManagementError, match='Unable to find management FQDN for host'):
            self.management.get_fqdn(hostname)



"""MediaWiki module."""
import logging

from typing import Any, Dict, Tuple

import requests

from cumin.transports import Command

from spicerack.confctl import ConftoolEntity
from spicerack.constants import CORE_DATACENTERS
from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError
from spicerack.remote import Remote, RemoteExecutionError, RemoteHosts


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class MediaWikiError(SpicerackError):
    """Custom exception class for errors in this module."""


class MediaWikiCheckError(SpicerackCheckError):
    """Custom exception class for checking errors in this module."""


class MediaWiki:
    """Class to manage MediaWiki-specific resources."""

    _list_cronjobs_command: str = '"$(crontab -u www-data -l | sed -r \'/^(#|$)/d\')"'
    _siteinfo_url: str = 'http://api.svc.{dc}.wmnet/w/api.php?action=query&meta=siteinfo&format=json&formatversion=2'

    def __init__(self, conftool: ConftoolEntity, remote: Remote, user: str, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            conftool (spicerack.confctl.ConftoolEntity): the conftool instance for the mwconfig type objects.
            remote (spicerack.remote.Remote): the Remote instance.
            user (str): the name of the effective running user.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._conftool = conftool
        self._remote = remote
        self._user = user
        self._dry_run = dry_run

    def check_config_line(self, filename: str, expected: str) -> bool:
        """Check that a MediaWiki configuration file contains some value.

        Arguments:
            filename (str): filename without extension in wmf-config.
            expected (str): string expected to be found in the configuration file.

        Returns:
            bool: True if the expected string is found in the configuration file, False otherwise.

        Raises:
            requests.exceptions.RequestException: on error.

        """
        noc_server = self._remote.query('O:Noc::Site').hosts[0]
        url = 'http://{noc}/conf/{filename}.php.txt'.format(noc=noc_server, filename=filename)
        mwconfig = requests.get(url, headers={'Host': 'noc.wikimedia.org'}, timeout=10)
        found = (expected in mwconfig.text)
        logger.debug('Checked message (found=%s) in MediaWiki config %s:\n%s', found, url, expected)

        return found

    @staticmethod
    def get_siteinfo(datacenter: str) -> Dict:
        """Get the JSON paylod for siteinfo from a random host in a given datacenter.

        Arguments:
            datacenter (str): the DC where to query for siteinfo.

        Returns:
            dict: the parsed JSON from siteinfo.

        Raises:
            requests.exceptions.RequestException: on failure.

        """
        url = MediaWiki._siteinfo_url.format(dc=datacenter)
        headers = {'X-Forwarded-Proto': 'https', 'Host': 'en.wikipedia.org'}

        response = requests.get(url, headers=headers, timeout=3)
        response.raise_for_status()

        return response.json()

    @staticmethod
    def check_siteinfo(datacenter: str, checks: Dict[Tuple[str, ...], Any], samples: int = 1) -> None:
        """Check that a specific value in siteinfo matches the expected ones, on multiple hosts.

        Arguments:
            datacenter (str): the DC where to query for siteinfo.
            checks (dict): dictionary of items to check, in which the keys are tuples with the path of keys to traverse
                the siteinfo dictionary to get the value and the values are the expected values to check. To check
                ``siteinfo[key1][key2]`` for a value ``value``, use::

                    {('key1', 'key2'): 'value'}

            samples (int, optional): the number of different calls to siteinfo to perform.

        Raises:
            MediaWikiError: if unable to get siteinfo or unable to traverse the siteinfo dictionary after all tries.
            MediaWikiCheckError: if the value doesn't match after all tries.

        """
        for i in range(1, samples + 1):  # Randomly check different hosts from the load balancer
            logger.debug('Checking siteinfo %d/%d', i, samples)
            MediaWiki._check_siteinfo(datacenter, checks)

    def scap_sync_config_file(self, filename: str, message: str) -> None:
        """Execute scap sync-file to deploy a specific configuration file of wmf-config.

        Arguments:
            filename (str): the filename without extension of wmf-config.
            message (str): the message to use for the scap sync-file execution.

        Raises:
            spicerack.remote.RemoteExecutionError: on error.

        """
        logger.debug('Syncing MediaWiki wmf-config/%s.php', filename)
        query = 'C:Deployment::Rsync and R:Class%cron_ensure = absent'
        command = 'su - {user} -c \'scap sync-file --force wmf-config/{filename}.php "{message}"\''.format(
            user=self._user, filename=filename, message=message)
        self._remote.query(query).run_sync(command)

    def set_readonly(self, datacenter: str, message: str) -> None:
        """Set the Conftool readonly variable for MediaWiki config in a specific datacenter.

        Arguments:
            datacenter (str): the DC for which the configuration must be changed.
            message (str): the readonly message string to set in MediaWiki.

        Raises:
            spicerack.confctl.ConfctlError: on Conftool errors and failed validation.
            spicerack.mediawiki.MediaWikiError: on failed siteinfo validation.

        """
        self._conftool.set_and_verify('val', message, scope=datacenter, name='ReadOnly')
        self._check_siteinfo_dry_run_aware(datacenter, {
            ('query', 'general', 'readonly'): True,
            ('query', 'general', 'readonlyreason'): message}, samples=10)

    def set_readwrite(self, datacenter: str) -> None:
        """Set the Conftool readonly variable for MediaWiki config to False to make it read-write.

        Arguments:
            datacenter (str): the DC for which the configuration must be changed.

        Raises:
            spicerack.confctl.ConfctlError: on Conftool errors and failed validation.
            spicerack.mediawiki.MediaWikiError: on failed siteinfo validation.

        """
        self._conftool.set_and_verify('val', False, scope=datacenter, name='ReadOnly')
        self._check_siteinfo_dry_run_aware(datacenter, {('query', 'general', 'readonly'): False}, samples=10)

    def set_master_datacenter(self, datacenter: str) -> None:
        """Set the MediaWiki config master datacenter variable in Conftool.

        Arguments:
            datacenter (str): the new master datacenter.

        Raises:
            spicerack.confctl.ConfctlError: on error.

        """
        self._conftool.set_and_verify('val', datacenter, scope='common', name='WMFMasterDatacenter')
        for dc in CORE_DATACENTERS:
            self._check_siteinfo_dry_run_aware(dc, {
                ('query', 'general', 'wmf-config', 'wmfMasterDatacenter'): datacenter}, samples=10)

    def get_maintenance_host(self, datacenter: str) -> RemoteHosts:
        """Get an instance to execute commands on the maintenance hosts in a given datacenter.

        Arguments:
            datacenter (str): the datacenter to filter for.

        Returns:
            spicerack.remote.RemoteHosts: the instance for the target host.

        """
        return self._remote.query('A:mw-maintenance and A:' + datacenter)

    def check_cronjobs_enabled(self, datacenter: str) -> None:
        """Check that MediaWiki cronjobs are set in the given DC.

        Arguments:
            datacenter (str): the name of the datacenter to work on.

        Raises:
            spicerack.remote.RemoteExecutionError: on failure.

        """
        self.get_maintenance_host(datacenter).run_sync('test ' + MediaWiki._list_cronjobs_command, is_safe=True)

    def check_cronjobs_disabled(self, datacenter: str) -> None:
        """Check that MediaWiki cronjobs are not set in the given DC.

        Arguments:
            datacenter (str): the name of the datacenter to work on.

        Raises:
            spicerack.remote.RemoteExecutionError: on failure.

        """
        self.get_maintenance_host(datacenter).run_sync('test -z ' + MediaWiki._list_cronjobs_command, is_safe=True)

    def stop_cronjobs(self, datacenter: str) -> None:
        """Remove and ensure MediaWiki cronjobs are not present in the given DC.

        Arguments:
            datacenter (str): the name of the datacenter to work on.

        Raises:
            spicerack.remote.RemoteExecutionError: on failure.

        """
        targets = self.get_maintenance_host(datacenter)
        logger.info('Disabling MediaWiki cronjobs in %s', datacenter)

        pkill_ok_codes = [0, 1]  # Accept both matches and no matches
        targets.run_async(
            Command('crontab -u www-data -r', ok_codes=[]),  # Cleanup the crontab
            # Kill all processes created by CRON for the www-data user
            Command('pkill -U www-data sh', ok_codes=pkill_ok_codes),
            # Kill MediaWiki wrappers, see modules/scap/manifests/scripts.pp in the Puppet repo
            Command('pkill --full "/usr/local/bin/foreachwiki"', ok_codes=pkill_ok_codes),
            Command('pkill --full "/usr/local/bin/foreachwikiindblist"', ok_codes=pkill_ok_codes),
            Command('pkill --full "/usr/local/bin/expanddblist"', ok_codes=pkill_ok_codes),
            Command('pkill --full "/usr/local/bin/mwscript"', ok_codes=pkill_ok_codes),
            Command('pkill --full "/usr/local/bin/mwscriptwikiset"', ok_codes=pkill_ok_codes),
            Command('killall -r php', ok_codes=[]),  # Kill all remaining PHP processes for all users
            'sleep 5',
            Command('killall -9 -r php', ok_codes=[]),  # No more time to be gentle
            'sleep 1',
            Command('systemctl start php7.2-fpm'),  # Restart the PHP-FPM services that killed above

        )
        self.check_cronjobs_disabled(datacenter)

        try:
            targets.run_sync('! pgrep -c php', is_safe=True)
        except RemoteExecutionError:
            # We just log an error, don't actually report a failure to the system. We can live with this.
            logger.error('Stray php processes still present on the maintenance host, please check')

    @staticmethod
    @retry(tries=5, backoff_mode='constant', exceptions=(MediaWikiError, MediaWikiCheckError))
    def _check_siteinfo(datacenter: str, checks: Dict[Tuple[str], Any]) -> None:
        """Check that a specific value in siteinfo matches the expected ones, retrying if doesn't match.

        Arguments:
            datacenter (str): the DC where to query for siteinfo.
            checks (dict): dictionary of items to check, in which the keys are tuples with the path of keys to traverse
                the siteinfo dictionary to get the value and the values are the expected values to check. To check
                ``siteinfo[key1][key2]`` for a value ``value``, use::

                    {('key1', 'key2'): 'value'}

        Raises:
            MediaWikiError: if unable to get siteinfo or unable to traverse the siteinfo dictionary after all tries.
            MediaWikiCheckError: if the value doesn't match after all tries.

        """
        try:
            siteinfo = MediaWiki.get_siteinfo(datacenter)
        except Exception as e:
            raise MediaWikiError('Failed to get siteinfo') from e

        for path, expected in checks.items():
            value = siteinfo.copy()  # No need for deepcopy, it will not be modified
            for key in path:
                try:
                    value = value[key]
                except (KeyError, TypeError) as e:
                    raise MediaWikiError('Failed to traverse siteinfo for key {key}'.format(key=key)) from e

            if value != expected:
                raise MediaWikiCheckError("Expected '{expected}', got '{value}' for path: {path}".format(
                    expected=expected, value=value, path=path))

    def _check_siteinfo_dry_run_aware(
        self,
        datacenter: str,
        checks: Dict[Tuple[str, ...], Any],
        samples: int = 1
    ) -> None:
        """Dry-run mode aware check_siteinfo. See check_siteinfo() documentation for more details.

        Arguments:
            datacenter (str): the DC where to query for siteinfo.
            checks (dict): dictionary of items to check, in which the keys are tuples with the path of keys to traverse
                the siteinfo dictionary to get the value and the values are the expected values to check. To check
                ``siteinfo[key1][key2]`` for a value ``value``, use::

                    {('key1', 'key2'): 'value'}

            samples (int, optional): the number of different calls to siteinfo to perform.

        Raises:
            MediaWikiError: if unable to get siteinfo or unable to traverse the siteinfo dictionary after all tries.
            MediaWikiCheckError: if the value doesn't match after all tries.

        """
        if self._dry_run:
            logger.debug('Reset samples to check_siteinfo from %s to 1 in dry-run mode', samples)
            samples = 1

        try:
            self.check_siteinfo(datacenter, checks, samples=samples)
        except (MediaWikiError, MediaWikiCheckError) as e:
            if self._dry_run:
                logger.info(e)
            else:
                raise



"""Tests package for Spicerack."""
import os

import pytest

from pkg_resources import get_distribution, parse_version


TESTS_BASE_PATH = os.path.realpath(os.path.dirname(__file__))
CAPLOG_MIN_VERSION = '3.3.0'
REQUESTS_MOCK_MIN_VERSION = '1.5.0'
ELASTICSEARCH_MIN_VERSION = '5.0.0'


def get_fixture_path(*paths):
    """Return the absolute path of the given fixture.

    Arguments:
        *paths: arbitrary positional arguments used to compose the absolute path to the fixture.

    Returns:
        str: the absolute path of the selected fixture.

    """
    return os.path.join(TESTS_BASE_PATH, 'fixtures', *paths)


SPICERACK_TEST_PARAMS = {
    'cumin_config': get_fixture_path('remote', 'config.yaml'),
    'conftool_config': get_fixture_path('confctl', 'config.yaml'),
    'conftool_schema': get_fixture_path('confctl', 'schema.yaml'),
    'debmonitor_config': get_fixture_path('debmonitor', 'config.ini'),
    'spicerack_config_dir': get_fixture_path(),
}

require_requests_mock = pytest.mark.skipif(  # pylint: disable=invalid-name
    parse_version(get_distribution('requests_mock').version) < parse_version(REQUESTS_MOCK_MIN_VERSION),
    reason='Requires requests-mock fixture')

require_caplog = pytest.mark.skipif(  # pylint: disable=invalid-name
    parse_version(pytest.__version__) < parse_version(CAPLOG_MIN_VERSION), reason='Requires caplog fixture')

min_elasticsearch = pytest.mark.skipif(  # pylint: disable=invalid-name
    parse_version(get_distribution('elasticsearch').version) < parse_version(ELASTICSEARCH_MIN_VERSION),
    reason='Requires more recent elasticsearch module')



"""Interactive module tests."""
from unittest import mock

import pytest

from spicerack import interactive
from spicerack.exceptions import SpicerackError


@mock.patch('builtins.input')
@mock.patch('spicerack.interactive.sys.stdout.isatty')
def test_ask_confirmation_ok(mocked_isatty, mocked_input, capsys):
    """Calling ask_confirmation() should not raise if the correct answer is provided."""
    mocked_isatty.return_value = True
    mocked_input.return_value = 'done'
    message = 'Test message'
    interactive.ask_confirmation(message)
    out, _ = capsys.readouterr()
    assert message in out


@mock.patch('builtins.input')
@mock.patch('spicerack.interactive.sys.stdout.isatty')
def test_ask_confirmation_ko(mocked_isatty, mocked_input, capsys):
    """Calling ask_confirmation() should raise if the correct answer is not provided."""
    mocked_isatty.return_value = True
    mocked_input.return_value = 'invalid'
    message = 'Test message'
    with pytest.raises(SpicerackError, match='Too many invalid confirmation answers'):
        interactive.ask_confirmation(message)

    out, _ = capsys.readouterr()
    assert message in out
    assert out.count('Invalid response') == 3


@mock.patch('spicerack.interactive.sys.stdout.isatty')
def test_ask_confirmation_no_tty(mocked_isatty):
    """It should raise SpicerackError if not in a TTY."""
    mocked_isatty.return_value = False
    with pytest.raises(SpicerackError, match='Not in a TTY, unable to ask for confirmation'):
        interactive.ask_confirmation('message')


def test_get_username_no_env(monkeypatch):
    """If no env variable is set should return '-'."""
    monkeypatch.delenv('USER', raising=False)
    monkeypatch.delenv('SUDO_USER', raising=False)
    assert interactive.get_username() == '-'


def test_get_username_root(monkeypatch):
    """When unable to detect the real user should return 'root'."""
    monkeypatch.setenv('USER', 'root')
    monkeypatch.delenv('SUDO_USER', raising=False)
    assert interactive.get_username() == 'root'


def test_get_username_ok(monkeypatch):
    """As a normal user with sudo should return the user's name."""
    monkeypatch.setenv('USER', 'root')
    monkeypatch.setenv('SUDO_USER', 'user')
    assert interactive.get_username() == 'user'


@mock.patch('spicerack.interactive.sys.stdout.isatty')
def test_ensure_shell_is_durable_interactive(mocked_isatty):
    """Should raise SpicerackError if in an interactive shell."""
    mocked_isatty.return_value = True
    with pytest.raises(SpicerackError, match='Must be run in non-interactive mode or inside a screen or tmux.'):
        interactive.ensure_shell_is_durable()

    assert mocked_isatty.called


@mock.patch('spicerack.interactive.sys.stdout.isatty')
def test_ensure_shell_is_durable_non_interactive(mocked_isatty):
    """Should raise SpicerackError if in an interactive shell."""
    mocked_isatty.return_value = False
    interactive.ensure_shell_is_durable()
    assert mocked_isatty.called


@mock.patch('spicerack.interactive.sys.stdout.isatty')
@pytest.mark.parametrize('env_name, env_value', (
    ('STY', '12345.pts-1.host'),
    ('TMUX', '/tmux-1001/default,12345,0'),
    ('TERM', 'screen-example'),
))
def test_ensure_shell_is_durable_sty(mocked_isatty, env_name, env_value, monkeypatch):
    """Should not raise if in an interactive shell with STY set, TMUX set or a screen-line TERM."""
    mocked_isatty.return_value = True
    monkeypatch.setenv(env_name, env_value)
    interactive.ensure_shell_is_durable()
    assert mocked_isatty.called


def test_get_management_password_in_env(monkeypatch):
    """Should return the MGMT_PASSWORD environment variable value, if set."""
    monkeypatch.setenv('MGMT_PASSWORD', 'env_password')
    assert interactive.get_management_password() == 'env_password'


@mock.patch('spicerack.interactive.getpass')
def test_get_management_password_interactive(mocked_getpass, monkeypatch):
    """Should ask for the password the MGMT_PASSWORD environment variable value, if set."""
    monkeypatch.delenv('MGMT_PASSWORD', raising=False)
    mocked_getpass.getpass.return_value = 'interactive_password'
    assert interactive.get_management_password() == 'interactive_password'
    mocked_getpass.getpass.assert_called_once_with(prompt='Management Password: ')


def test_get_management_password_empty(monkeypatch):
    """Should raise SpicerackError if the password is empty."""
    monkeypatch.setenv('MGMT_PASSWORD', '')
    with pytest.raises(SpicerackError, match='Empty Management Password'):
        interactive.get_management_password()


@mock.patch('spicerack.interactive.getpass')
def test_get_secret_correct_noconfirm(mocked_getpass):
    """Should ask for secret once and return the secret."""
    mocked_getpass.getpass.return_value = 'interactive_password'
    assert interactive.get_secret('secret') == 'interactive_password'
    mocked_getpass.getpass.assert_called_once_with(prompt='secret: ')


@mock.patch('spicerack.interactive.getpass')
def test_get_secret_correct(mocked_getpass):
    """Should ask for secret twice and return the secret."""
    mocked_getpass.getpass.side_effect = ['interactive_password', 'interactive_password']
    assert interactive.get_secret('secret', confirm=True) == 'interactive_password'
    mocked_getpass.getpass.assert_has_calls(
        [mock.call(prompt='secret: '), mock.call(prompt='Again, just to be sure: ')])


@mock.patch('spicerack.interactive.getpass')
def test_get_secret_bad_retry(mocked_getpass):
    """Should ask for secret twice and raise SpicerackError if they don't match."""
    mocked_getpass.getpass.side_effect = ['interactive_password', 'foobar']
    with pytest.raises(SpicerackError, match='secret: Passwords did not match'):
        interactive.get_secret('secret', confirm=True)
    mocked_getpass.getpass.assert_has_calls(
        [mock.call(prompt='secret: '), mock.call(prompt='Again, just to be sure: ')])


@mock.patch('spicerack.interactive.getpass')
def test_get_secret_too_small(mocked_getpass):
    """Should ask for secret until the minimum length is met."""
    mocked_getpass.getpass.side_effect = ['5char', 'interactive_password']
    assert interactive.get_secret('secret') == 'interactive_password'
    mocked_getpass.getpass.assert_has_calls(
        [mock.call(prompt='secret: '),
         mock.call(prompt='Secret must be at least 6 characters. try again: ')])



"""Phabricator module."""
import configparser
import logging

import phabricator

from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def create_phabricator(
    bot_config_file: str,
    section: str = 'phabricator_bot',
    dry_run: bool = True
) -> phabricator.Phabricator:
    """Initialize the Phabricator client from the bot config file.

    Arguments:
        bot_config_file (str): the path to the configuration file for the Phabricator bot, with the following
            structure::

                [section_name]
                host = https://phabricator.example.com/api/
                username = phab-bot
                token = api-12345

        section (str, optional): the name of the section of the configuration file where to find the required
            parameters.
        dry_run (bool, optional): whether this is a DRY-RUN.

    Returns:
        spicerack.phabricator.Phabricator: a Phabricator instance.

    Raises:
        spicerack.phabricator.PhabricatorError: if unable to get all the required parameters from the bot configuration
            file, or to initialize the Phabricator client.

    """
    parser = configparser.ConfigParser()
    parser.read(bot_config_file)
    required_options = ('host', 'username', 'token')
    params = {}

    try:
        for option in required_options:
            params[option] = parser.get(section, option)
    except configparser.NoSectionError as e:
        raise PhabricatorError('Unable to find section {section} in config file {file}'.format(
            section=section, file=bot_config_file)) from e
    except configparser.NoOptionError as e:
        raise PhabricatorError(
            'Unable to find all required options {options} in section {section} of config file {file}'.format(
                options=required_options, section=section, file=bot_config_file)) from e

    try:
        client = phabricator.Phabricator(**params)
    except Exception as e:
        raise PhabricatorError('Unable to instantiate Phabricator client') from e

    return Phabricator(client, dry_run=dry_run)


class PhabricatorError(SpicerackError):
    """Custom exception class for errors of the Phabricator class."""


class Phabricator:
    """Class to interact with a Phabricator website."""

    def __init__(self, phabricator_client: phabricator.Phabricator, dry_run: bool = True) -> None:
        """Initialize the Phabricator client from the bot config file.

        Arguments:
            phabricator_client (phabricator.Phabricator): a Phabricator client instance.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._client = phabricator_client
        self._dry_run = dry_run

    def task_comment(self, task_id: str, comment: str) -> None:
        """Add a comment on a Phabricator task.

        Arguments:
            task_id (str): the Phabricator task ID (e.g. ``T12345``) to be updated.
            comment (str): the message to add to the task.

        Raises:
            spicerack.phabricator.PhabricatorError: if unable to update the task.

        """
        if self._dry_run:
            logger.debug('Skip updating Phabricator task %s in DRY-RUN with comment: %s', task_id, comment)
            return

        try:
            transactions = [{'type': 'comment', 'value': comment}]
            self._client.maniphest.edit(objectIdentifier=task_id, transactions=transactions)
            logger.info('Updated Phabricator task %s', task_id)
        except Exception as e:
            raise PhabricatorError('Unable to update Phabricator task {id}'.format(id=task_id)) from e



"""Netbox module tests."""

from unittest import mock

import pynetbox
import pytest

from spicerack.netbox import Netbox, NetboxAPIError, NetboxError, NetboxHostNotFoundError


NETBOX_URL = 'https://example.com/'
NETBOX_TOKEN = 'FAKETOKEN'  # nosec


def _request_error():
    """Return a Netbox RequestError."""
    fakestatus = mock.Mock()
    fakestatus.status_code = 404
    return pynetbox.RequestError(fakestatus)


def _fake_host():
    """Return a mocked Netbox host."""
    fake_host = mock.Mock()
    fake_host.name = 'test'
    fake_host.status = 'Active'
    fake_host.serialize = mock.Mock(return_value={'name': 'test'})
    fake_host.save = mock.Mock(return_value=True)
    return fake_host


def _fake_virtual_host():
    """Return a mocked Netbox host."""
    fake_host = mock.Mock()
    fake_host.name = 'test'
    fake_host.status = 'Active'
    fake_host.cluster.name = 'testcluster'
    fake_host.serialize = mock.Mock(return_value={'name': 'test'})
    fake_host.save = mock.Mock(return_value=True)
    return fake_host


@mock.patch('pynetbox.api')
def test_netbox_api(mocked_api):
    """An instance of Netbox should instantiate the Netbox API and expose it via the api property."""
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)
    mocked_api.called_once_with(NETBOX_URL, token=NETBOX_TOKEN)
    assert netbox.api == mocked_api()


@mock.patch('pynetbox.api')
def test_netbox_fetch_host_status_nohost(mocked_pynetbox):
    """Test the error scenario where the host is not found."""
    mocked_pynetbox().dcim.devices.get.return_value = None
    mocked_pynetbox().virtualization.virtual_machines.get.return_value = None
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)
    with pytest.raises(NetboxHostNotFoundError):
        netbox.fetch_host_status('host')


@mock.patch('pynetbox.api')
def test_netbox_fetch_host_status_error(mocked_pynetbox):
    """Test the error scenario where the host is not found."""
    mocked_pynetbox().dcim.devices.get = mock.Mock(side_effect=_request_error())
    mocked_pynetbox().virtualization.virtual_machines.get = mock.Mock(side_effect=_request_error())
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)
    with pytest.raises(NetboxError, match='Error retrieving Netbox host'):
        netbox.fetch_host_status('host')


@mock.patch('pynetbox.api')
def test_fetch_host_status(mocked_pynetbox):
    """Test fetching host status."""
    mocked_pynetbox().dcim.devices.get.return_value = _fake_host()
    mocked_pynetbox().virtualization.virtual_machines.get.return_value = None
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)

    assert netbox.fetch_host_status('host') == 'Active'


@mock.patch('pynetbox.api')
def test_put_host_status_good_status(mocked_pynetbox):
    """Test setting a status and it working."""
    fake_host = _fake_host()
    mocked_pynetbox().dcim.devices.get.return_value = fake_host
    mocked_pynetbox().virtualization.virtual_machines.get.return_value = None
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=False)

    netbox.put_host_status('host', 'Planned')
    assert str(fake_host.status) == 'planned'
    assert fake_host.save.called


@mock.patch('pynetbox.api')
def test_put_host_status_save_failure(mocked_pynetbox):
    """Test save failure."""
    fake_host = _fake_host()
    mocked_pynetbox().dcim.devices.get.return_value = fake_host
    mocked_pynetbox().virtualization.virtual_machines.get.return_value = None
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=False)
    fake_host.save.return_value = False
    with pytest.raises(NetboxAPIError, match='Failed to update Netbox status for host testhost Active -> planned'):
        netbox.put_host_status('testhost', 'Planned')
    assert fake_host.save.called


@mock.patch('pynetbox.api')
def test_put_host_status_dryrun_success(mocked_pynetbox):
    """Test dry run, which  should always work on save."""
    fake_host = _fake_host()
    mocked_pynetbox().dcim.devices.get.return_value = fake_host
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)
    fake_host.save.return_value = False
    netbox.put_host_status('host', 'Planned')
    assert not fake_host.save.called


@mock.patch('pynetbox.api')
def test_put_host_status_error(mocked_pynetbox):
    """Test exception during save."""
    fake_host = _fake_host()
    mocked_pynetbox().dcim.devices.get.return_value = fake_host
    mocked_pynetbox().virtualization.virtual_machines.get.return_value = None
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=False)
    fake_host.save.side_effect = _request_error()
    with pytest.raises(NetboxAPIError, match='Failed to save Netbox status for host testhost Active -> planned'):
        netbox.put_host_status('testhost', 'Planned')
    assert fake_host.save.called


@mock.patch('pynetbox.api')
def test_fetch_host_detail(mocked_pynetbox):
    """Test fetching host detail."""
    fake_host = _fake_host()
    mocked_pynetbox().dcim.devices.get.return_value = fake_host
    mocked_pynetbox().virtualization.virtual_machines.get.return_value = None
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)
    detail = netbox.fetch_host_detail('test')
    assert fake_host.serialize.called
    assert not detail['is_virtual']


@mock.patch('pynetbox.api')
def test_fetch_host_detail_vm(mocked_pynetbox):
    """Virtual machines should have is_virtual == True and a cluster name."""
    fake_host = _fake_virtual_host()
    mocked_pynetbox().dcim.devices.get.return_value = None
    mocked_pynetbox().virtualization.virtual_machines.get.return_value = fake_host
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)
    detail = netbox.fetch_host_detail('test')
    assert fake_host.serialize.called
    assert detail['is_virtual']
    assert detail['ganeti_cluster'] == 'testcluster'


@mock.patch('pynetbox.api')
def test_fetch_host_status_vm(mocked_pynetbox):
    """Virtual machines should return status just like devices."""
    mocked_pynetbox().dcim.devices.get.return_value = None
    mocked_pynetbox().virtualization.virtual_machines.get.return_value = _fake_virtual_host()
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)

    assert netbox.fetch_host_status('host') == 'Active'


@mock.patch('pynetbox.api')
def test_fetch_host_status_vm_error(mocked_pynetbox):
    """Virtual machines should raise an exception on an API error."""
    mocked_pynetbox().dcim.devices.get.return_value = None
    mocked_pynetbox().virtualization.virtual_machines.get = mock.Mock(side_effect=_request_error())
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)

    with pytest.raises(NetboxAPIError, match="Error retrieving Netbox VM"):
        netbox.fetch_host_status('host')


@mock.patch('pynetbox.api')
def test_set_host_status_vm(mocked_pynetbox):
    """Virtual machines should raise an exception if you try to set the status."""
    fake_vm = _fake_virtual_host()
    mocked_pynetbox().dcim.devices.get.return_value = None
    mocked_pynetbox().virtualization.virtual_machines.get.return_value = fake_vm
    netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)

    with pytest.raises(NetboxHostNotFoundError):
        netbox.put_host_status('test', 'Active')

    assert not fake_vm.save.called



"""Administrative module tests."""
import pytest

from spicerack import administrative


def test_reason_init():
    """It should initalize a Reason instance with the mandatory parameters."""
    reason = administrative.Reason('Reason message', 'user1', 'host1')
    assert str(reason) == 'Reason message - user1@host1'


def test_reason_init_with_task():
    """It should initalize a Reason instance with the optional task ID."""
    reason = administrative.Reason('Reason message', 'user1', 'host1', task_id='T12345')
    assert str(reason) == 'Reason message - user1@host1 - T12345'


@pytest.mark.parametrize('failing_param_name, args, kwargs', (
    ('_reason', ['Reason with "double quotes"', 'user1', 'host1'], {'task_id': ''}),
    ('_username', ['Reason message', 'user"', 'host1'], {'task_id': ''}),
    ('_hostname', ['Reason message', 'user1', 'host1"'], {'task_id': ''}),
    ('_task_id', ['Reason message', 'user1', 'host1'], {'task_id': 'T"'}),
))
def test_reason_init_double_quotes_in_params(failing_param_name, args, kwargs):
    """It should raise ReasonError if any of the parameters contains double quotes."""
    with pytest.raises(administrative.ReasonError, match='Property {name} cannot contain double quotes'.format(
            name=failing_param_name)):
        administrative.Reason(*args, **kwargs)


def test_reason_owner():
    """It should return the owner part."""
    reason = administrative.Reason('Reason message', 'user1', 'host1')
    assert reason.owner == 'user1@host1'


def test_reason_hostname():
    """It should return the hostname."""
    reason = administrative.Reason('Reason message', 'user1', 'host1')
    assert reason.hostname == 'host1'


def test_reason_quoted():
    """It should return the double quoted string representation of the instance."""
    reason = administrative.Reason('Reason message', 'user1', 'host1')
    assert reason.quoted() == '"Reason message - user1@host1"'
